import numpy as np
import openmdao.api as om

from wisdem.commonse import gravity, eps, NFREQ
import wisdem.commonse.frustum as frustum
import wisdem.commonse.manufacturing as manufacture
from wisdem.commonse.utilization_constraints import hoopStressEurocode, hoopStress
import wisdem.commonse.utilities as util
import wisdem.pyframe3dd.pyframe3dd as pyframe3dd


RIGID = 1e30
NREFINE = 3


def get_nfull(npts):
    nFull = int(1 + NREFINE * (npts - 1))
    return nFull


def get_npts(nFull):
    npts = int(1 + (nFull - 1) / NREFINE)
    return npts


# -----------------
#  Components
# -----------------

# TODO need to check the length of each array
class CylinderDiscretization(om.ExplicitComponent):
    """
    Discretize geometry into finite element nodes

    Parameters
    ----------
    foundation_height : float, [m]
        starting height of tower
    section_height : numpy array[nPoints-1], [m]
        parameterized section heights along cylinder
    diameter : numpy array[nPoints], [m]
        cylinder diameter at corresponding locations
    wall_thickness : numpy array[nPoints-1], [m]
        shell thickness at corresponding locations

    Returns
    -------
    z_param : numpy array[nPoints], [m]
        parameterized locations along cylinder, linear lofting between
    z_full : numpy array[nFull], [m]
        locations along cylinder
    d_full : numpy array[nFull], [m]
        cylinder diameter at corresponding locations
    t_full : numpy array[nFull-1], [m]
        shell thickness at corresponding locations

    """

    """discretize geometry into finite element nodes"""

    def initialize(self):
        self.options.declare("nPoints")
        self.options.declare("nRefine", default=NREFINE)

    def setup(self):
        nPoints = self.options["nPoints"]
        nRefine = np.round(self.options["nRefine"])
        nFull = int(nRefine * (nPoints - 1) + 1)

        self.add_input("foundation_height", val=0.0, units="m")
        self.add_input("section_height", np.zeros(nPoints - 1), units="m")
        self.add_input("diameter", np.zeros(nPoints), units="m")
        self.add_input("wall_thickness", np.zeros(nPoints - 1), units="m")

        self.add_output("z_param", np.zeros(nPoints), units="m")
        self.add_output("z_full", np.zeros(nFull), units="m")
        self.add_output("d_full", np.zeros(nFull), units="m")
        self.add_output("t_full", np.zeros(nFull - 1), units="m")

        # self.declare_partials('*', '*', method='fd', form='central', step=1e-6)

    def compute(self, inputs, outputs):
        # Check to make sure we have good values
        if np.any(inputs["section_height"] <= 0.0):
            raise ValueError("Section height values must be greater than zero, " + str(inputs["section_height"]))
        if np.any(inputs["wall_thickness"] <= 0.0):
            raise ValueError("Wall thickness values must be greater than zero, " + str(inputs["wall_thickness"]))
        if np.any(inputs["diameter"] <= 0.0):
            raise ValueError("Diameter values must be greater than zero, " + str(inputs["diameter"]))

        nRefine = int(np.round(self.options["nRefine"]))
        z_param = float(inputs["foundation_height"]) + np.r_[0.0, np.cumsum(inputs["section_height"].flatten())]
        # Have to regine each element one at a time so that we preserve input nodes
        z_full = np.array([])
        for k in range(z_param.size - 1):
            zref = np.linspace(z_param[k], z_param[k + 1], nRefine + 1)
            z_full = np.append(z_full, zref)
        z_full = np.unique(z_full)
        outputs["z_full"] = z_full
        outputs["d_full"] = np.interp(z_full, z_param, inputs["diameter"])
        z_section = 0.5 * (z_full[:-1] + z_full[1:])
        outputs["t_full"] = util.sectionalInterp(z_section, z_param, inputs["wall_thickness"])
        outputs["z_param"] = z_param


class CylinderMass(om.ExplicitComponent):
    """
    Compute cylinder cost and mass properties

    Parameters
    ----------
    d_full : numpy array[nPoints], [m]
        cylinder diameter at corresponding locations
    t_full : numpy array[nPoints-1], [m]
        shell thickness at corresponding locations
    z_full : numpy array[nPoints], [m]
        parameterized locations along cylinder, linear lofting between
    rho : numpy array[nPoints-1], [kg/m**3]
        material density
    outfitting_factor : numpy array[nPoints-1]
        Multiplier that accounts for secondary structure mass inside of cylinder
    material_cost_rate : numpy array[nPoints-1], [USD/kg]
        Raw material cost rate: steel $1.1/kg, aluminum $3.5/kg
    labor_cost_rate : float, [USD/min]
        Labor cost rate
    painting_cost_rate : float, [USD/m/m]
        Painting / surface finishing cost rate

    Returns
    -------
    cost : float, [USD]
        Total cylinder cost
    mass : numpy array[nPoints-1], [kg]
        Total cylinder mass
    center_of_mass : float, [m]
        z-position of center of mass of cylinder
    section_center_of_mass : numpy array[nPoints-1], [m]
        z position of center of mass of each can in the cylinder
    I_base : numpy array[6], [kg*m**2]
        mass moment of inertia of cylinder about base [xx yy zz xy xz yz]

    """

    def initialize(self):
        self.options.declare("nPoints")

    def setup(self):
        nPoints = self.options["nPoints"]

        self.add_input("d_full", val=np.zeros(nPoints), units="m")
        self.add_input("t_full", val=np.zeros(nPoints - 1), units="m")
        self.add_input("z_full", val=np.zeros(nPoints), units="m")
        self.add_input("rho", val=np.zeros(nPoints - 1), units="kg/m**3")
        self.add_input("outfitting_factor", val=np.zeros(nPoints - 1))
        self.add_input("material_cost_rate", val=np.zeros(nPoints - 1), units="USD/kg")
        self.add_input("labor_cost_rate", 0.0, units="USD/min")
        self.add_input("painting_cost_rate", 0.0, units="USD/m/m")

        self.add_output("cost", val=0.0, units="USD")
        self.add_output("mass", val=np.zeros(nPoints - 1), units="kg")
        self.add_output("center_of_mass", val=0.0, units="m")
        self.add_output("section_center_of_mass", val=np.zeros(nPoints - 1), units="m")
        self.add_output("I_base", np.zeros(6), units="kg*m**2")

    def compute(self, inputs, outputs):
        # Unpack variables for thickness and average radius at each can interface
        twall = inputs["t_full"]
        Rb = 0.5 * inputs["d_full"][:-1]
        Rt = 0.5 * inputs["d_full"][1:]
        zz = inputs["z_full"]
        H = np.diff(zz)
        rho = inputs["rho"]
        coeff = inputs["outfitting_factor"]
        coeff = coeff + np.where(coeff < 1.0, 1.0, 0.0)

        # Total mass of cylinder
        V_shell = frustum.frustumShellVol(Rb, Rt, twall, H)
        mass = outputs["mass"] = coeff * rho * V_shell

        # Center of mass of each can/section
        cm_section = zz[:-1] + frustum.frustumShellCG(Rb, Rt, twall, H)
        outputs["section_center_of_mass"] = cm_section

        # Center of mass of cylinder
        V_shell += eps
        outputs["center_of_mass"] = np.dot(V_shell, cm_section) / V_shell.sum()

        # Moments of inertia
        Izz_section = coeff * rho * frustum.frustumShellIzz(Rb, Rt, twall, H)
        Ixx_section = Iyy_section = coeff * rho * frustum.frustumShellIxx(Rb, Rt, twall, H)

        # Sum up each cylinder section using parallel axis theorem
        I_base = np.zeros((3, 3))
        for k in range(Izz_section.size):
            R = np.array([0.0, 0.0, cm_section[k] - zz[0]])
            Icg = util.assembleI([Ixx_section[k], Iyy_section[k], Izz_section[k], 0.0, 0.0, 0.0])

            I_base += Icg + mass[k] * (np.dot(R, R) * np.eye(3) - np.outer(R, R))

        outputs["I_base"] = util.unassembleI(I_base)

        # Compute costs based on "Optimum Design of Steel Structures" by Farkas and Jarmai
        # All dimensions for correlations based on mm, not meters.
        R_ave = 0.5 * (Rb + Rt)
        taper = np.minimum(Rb / Rt, Rt / Rb)
        nsec = twall.size
        mshell = rho * V_shell
        mshell_tot = np.sum(rho * V_shell)
        k_m = inputs["material_cost_rate"]  # 1.1 # USD / kg carbon steel plate
        k_f = inputs["labor_cost_rate"]  # 1.0 # USD / min labor
        k_p = inputs["painting_cost_rate"]  # USD / m^2 painting
        k_e = 0.064  # Industrial electricity rate $/kWh https://www.eia.gov/electricity/monthly/epm_table_grapher.php?t=epmt_5_6_a
        e_f = 15.9  # Electricity usage kWh/kg for steel
        e_fo = 26.9  # Electricity usage kWh/kg for stainless steel

        # Cost Step 1) Cutting flat plates for taper using plasma cutter
        cutLengths = 2.0 * np.sqrt((Rt - Rb) ** 2.0 + H ** 2.0)  # Factor of 2 for both sides
        # Cost Step 2) Rolling plates
        # Cost Step 3) Welding rolled plates into shells (set difficulty factor based on tapering with logistic function)
        theta_F = 4.0 - 3.0 / (1 + np.exp(-5.0 * (taper - 0.75)))
        # Cost Step 4) Circumferential welds to join cans together
        theta_A = 2.0

        # Labor-based expenses
        K_f = k_f * (
            manufacture.steel_cutting_plasma_time(cutLengths, twall)
            + manufacture.steel_rolling_time(theta_F, R_ave, twall)
            + manufacture.steel_butt_welding_time(theta_A, nsec, mshell_tot, cutLengths, twall)
            + manufacture.steel_butt_welding_time(theta_A, nsec, mshell_tot, 2 * np.pi * Rb[1:], twall[1:])
        )

        # Cost step 5) Painting- outside and inside
        theta_p = 2
        K_p = k_p * theta_p * 2 * (2 * np.pi * R_ave * H).sum()

        # Cost step 6) Outfitting with electricity usage
        K_o = np.sum(1.5 * k_m * (coeff - 1.0) * mshell)

        # Material cost with waste fraction, but without outfitting,
        K_m = 1.21 * np.sum(k_m * mshell)

        # Electricity usage
        K_e = np.sum(k_e * (e_f * mshell + e_fo * (coeff - 1.0) * mshell))

        # Assemble all costs for now
        tempSum = K_m + K_e + K_o + K_p + K_f

        # Capital cost share from BLS MFP by NAICS
        K_c = 0.118 * tempSum / (1.0 - 0.118)

        outputs["cost"] = tempSum + K_c


# @implement_base(CylinderFromCSProps)
class CylinderFrame3DD(om.ExplicitComponent):
    """
    Run Frame3DD on the cylinder geometry

    Parameters
    ----------
    z : numpy array[npts], [m]
        location along cylinder. start at bottom and go to top
    Az : numpy array[npts-1], [m**2]
        cross-sectional area
    Asx : numpy array[npts-1], [m**2]
        x shear area
    Asy : numpy array[npts-1], [m**2]
        y shear area
    Jz : numpy array[npts-1], [m**4]
        polar moment of inertia
    Ixx : numpy array[npts-1], [m**4]
        area moment of inertia about x-axis
    Iyy : numpy array[npts-1], [m**4]
        area moment of inertia about y-axis
    E : numpy array[npts-1], [N/m**2]
        modulus of elasticity
    G : numpy array[npts-1], [N/m**2]
        shear modulus
    rho : numpy array[npts-1], [kg/m**3]
        material density
    d : numpy array[npts], [m]
        effective cylinder diameter for section
    t : numpy array[npts-1], [m]
        effective shell thickness for section
    kidx : numpy array[nK, dtype]
        indices of z where external stiffness reactions should be applied.
    kx : numpy array[nK], [N/m]
        spring stiffness in x-direction
    ky : numpy array[nK], [N/m]
        spring stiffness in y-direction
    kz : numpy array[nK], [N/m]
        spring stiffness in z-direction
    ktx : numpy array[nK], [N/m]
        spring stiffness in theta_x-rotation
    kty : numpy array[nK], [N/m]
        spring stiffness in theta_y-rotation
    ktz : numpy array[nK], [N/m]
        spring stiffness in theta_z-rotation
    midx : numpy array[nMass, dtype]
        indices where added mass should be applied.
    m : numpy array[nMass], [kg]
        added mass
    mIxx : numpy array[nMass], [kg*m**2]
        x mass moment of inertia about some point p
    mIyy : numpy array[nMass], [kg*m**2]
        y mass moment of inertia about some point p
    mIzz : numpy array[nMass], [kg*m**2]
        z mass moment of inertia about some point p
    mIxy : numpy array[nMass], [kg*m**2]
        xy mass moment of inertia about some point p
    mIxz : numpy array[nMass], [kg*m**2]
        xz mass moment of inertia about some point p
    mIyz : numpy array[nMass], [kg*m**2]
        yz mass moment of inertia about some point p
    mrhox : numpy array[nMass], [m]
        x-location of p relative to node
    mrhoy : numpy array[nMass], [m]
        y-location of p relative to node
    mrhoz : numpy array[nMass], [m]
        z-location of p relative to node
    plidx : numpy array[nPL, dtype]
        indices where point loads should be applied.
    Fx : numpy array[nPL], [N]
        point force in x-direction
    Fy : numpy array[nPL], [N]
        point force in y-direction
    Fz : numpy array[nPL], [N]
        point force in z-direction
    Mxx : numpy array[nPL], [N*m]
        point moment about x-axis
    Myy : numpy array[nPL], [N*m]
        point moment about y-axis
    Mzz : numpy array[nPL], [N*m]
        point moment about z-axis
    Px : numpy array[npts], [N/m]
        force per unit length in x-direction
    Py : numpy array[npts], [N/m]
        force per unit length in y-direction
    Pz : numpy array[npts], [N/m]
        force per unit length in z-direction
    qdyn : numpy array[npts], [N/m**2]
        dynamic pressure

    Returns
    -------
    mass : float, [kg]
        Structural mass computed by Frame3DD
    f1 : float, [Hz]
        First natural frequency
    f2 : float, [Hz]
        Second natural frequency
    freqs : numpy array[NFREQ], [Hz]
        Natural frequencies of the structure
    x_mode_shapes : numpy array[NFREQ2, 5]
        6-degree polynomial coefficients of mode shapes in the x-direction
    y_mode_shapes : numpy array[NFREQ2, 5]
        6-degree polynomial coefficients of mode shapes in the x-direction
    top_deflection : float, [m]
        Deflection of cylinder top in yaw-aligned +x direction
    Fz_out : numpy array[npts-1], [N]
        Axial foce in vertical z-direction in cylinder structure.
    Vx_out : numpy array[npts-1], [N]
        Shear force in x-direction in cylinder structure.
    Vy_out : numpy array[npts-1], [N]
        Shear force in y-direction in cylinder structure.
    Mxx_out : numpy array[npts-1], [N*m]
        Moment about x-axis in cylinder structure.
    Myy_out : numpy array[npts-1], [N*m]
        Moment about y-axis in cylinder structure.
    Mzz_out : numpy array[npts-1], [N*m]
        Moment about z-axis in cylinder structure.
    base_F : numpy array[3], [N]
        Total force on cylinder
    base_M : numpy array[3], [N*m]
        Total moment on cylinder measured at base
    axial_stress : numpy array[npts-1], [N/m**2]
        Axial stress in cylinder structure
    shear_stress : numpy array[npts-1], [N/m**2]
        Shear stress in cylinder structure
    hoop_stress : numpy array[npts-1], [N/m**2]
        Hoop stress in cylinder structure calculated with simple method used in API
        standards
    hoop_stress_euro : numpy array[npts-1], [N/m**2]
        Hoop stress in cylinder structure calculated with Eurocode method

    """

    def initialize(self):
        self.options.declare("npts")
        self.options.declare("nK")
        self.options.declare("nMass")
        self.options.declare("nPL")
        self.options.declare("frame3dd_opt")
        self.options.declare("buckling_length")

    def setup(self):
        npts = self.options["npts"]
        nK = self.options["nK"]
        nMass = self.options["nMass"]
        nPL = self.options["nPL"]

        # cross-sectional data along cylinder.
        self.add_input("z", val=np.zeros(npts), units="m")
        self.add_input("Az", val=np.zeros(npts - 1), units="m**2")
        self.add_input("Asx", val=np.zeros(npts - 1), units="m**2")
        self.add_input("Asy", val=np.zeros(npts - 1), units="m**2")
        self.add_input("Jz", val=np.zeros(npts - 1), units="m**4")
        self.add_input("Ixx", val=np.zeros(npts - 1), units="m**4")
        self.add_input("Iyy", val=np.zeros(npts - 1), units="m**4")
        self.add_input("E", val=np.zeros(npts - 1), units="N/m**2")
        self.add_input("G", val=np.zeros(npts - 1), units="N/m**2")
        self.add_input("rho", val=np.zeros(npts - 1), units="kg/m**3")

        # effective geometry -- used for handbook methods to estimate hoop stress, buckling, fatigue
        self.add_input("d", val=np.zeros(npts), units="m")
        self.add_input("t", val=np.zeros(npts - 1), units="m")

        # spring reaction data.  Use global RIGID for rigid constraints.
        # JPJ: this next output should be a discrete output
        self.add_input("kidx", val=np.zeros(nK, dtype=np.int_))
        self.add_input("kx", val=np.zeros(nK), units="N/m")
        self.add_input("ky", val=np.zeros(nK), units="N/m")
        self.add_input("kz", val=np.zeros(nK), units="N/m")
        self.add_input("ktx", val=np.zeros(nK), units="N/m")
        self.add_input("kty", val=np.zeros(nK), units="N/m")
        self.add_input("ktz", val=np.zeros(nK), units="N/m")

        # extra mass
        # JPJ: this next output should be a discrete output
        self.add_input("midx", val=np.zeros(nMass, dtype=np.int_))
        self.add_input("m", val=np.zeros(nMass), units="kg")
        self.add_input("mIxx", val=np.zeros(nMass), units="kg*m**2")
        self.add_input("mIyy", val=np.zeros(nMass), units="kg*m**2")
        self.add_input("mIzz", val=np.zeros(nMass), units="kg*m**2")
        self.add_input("mIxy", val=np.zeros(nMass), units="kg*m**2")
        self.add_input("mIxz", val=np.zeros(nMass), units="kg*m**2")
        self.add_input("mIyz", val=np.zeros(nMass), units="kg*m**2")
        self.add_input("mrhox", val=np.zeros(nMass), units="m")
        self.add_input("mrhoy", val=np.zeros(nMass), units="m")
        self.add_input("mrhoz", val=np.zeros(nMass), units="m")

        # point loads (if addGravityLoadForExtraMass=True be sure not to double count by adding those force here also)
        self.add_input("plidx", val=np.zeros(nPL, dtype=np.int_))
        self.add_input("Fx", val=np.zeros(nPL), units="N")
        self.add_input("Fy", val=np.zeros(nPL), units="N")
        self.add_input("Fz", val=np.zeros(nPL), units="N")
        self.add_input("Mxx", val=np.zeros(nPL), units="N*m")
        self.add_input("Myy", val=np.zeros(nPL), units="N*m")
        self.add_input("Mzz", val=np.zeros(nPL), units="N*m")

        # combined wind-water distributed loads
        self.add_input("Px", val=np.zeros(npts), units="N/m")
        self.add_input("Py", val=np.zeros(npts), units="N/m")
        self.add_input("Pz", val=np.zeros(npts), units="N/m")
        self.add_input("qdyn", val=np.zeros(npts), units="N/m**2")

        NFREQ2 = int(NFREQ / 2)
        self.add_output("mass", val=0.0, units="kg")
        self.add_output("f1", val=0.0, units="Hz")
        self.add_output("f2", val=0.0, units="Hz")
        self.add_output("freqs", val=np.zeros(NFREQ), units="Hz")
        self.add_output("x_mode_shapes", val=np.zeros((NFREQ2, 5)))
        self.add_output("y_mode_shapes", val=np.zeros((NFREQ2, 5)))
        self.add_output("x_mode_freqs", val=np.zeros(NFREQ2))
        self.add_output("y_mode_freqs", val=np.zeros(NFREQ2))
        self.add_output("top_deflection", val=0.0, units="m")
        self.add_output("Fz_out", val=np.zeros(npts - 1), units="N")
        self.add_output("Vx_out", val=np.zeros(npts - 1), units="N")
        self.add_output("Vy_out", val=np.zeros(npts - 1), units="N")
        self.add_output("Mxx_out", val=np.zeros(npts - 1), units="N*m")
        self.add_output("Myy_out", val=np.zeros(npts - 1), units="N*m")
        self.add_output("Mzz_out", val=np.zeros(npts - 1), units="N*m")
        self.add_output("base_F", val=np.zeros(3), units="N")
        self.add_output("base_M", val=np.zeros(3), units="N*m")
        self.add_output("axial_stress", val=np.zeros(npts - 1), units="N/m**2")
        self.add_output("shear_stress", val=np.zeros(npts - 1), units="N/m**2")
        self.add_output("hoop_stress", val=np.zeros(npts - 1), units="N/m**2")
        self.add_output("hoop_stress_euro", val=np.zeros(npts - 1), units="N/m**2")

        # self.declare_partials('*', '*', method='fd', form='central', step=1e-6)

    def compute(self, inputs, outputs):

        frame3dd_opt = self.options["frame3dd_opt"]

        # ------- node data ----------------
        z = inputs["z"]
        n = len(z)
        node = np.arange(1, n + 1)
        x = np.zeros(n)
        y = np.zeros(n)
        r = np.zeros(n)

        nodes = pyframe3dd.NodeData(node, x, y, z, r)
        # -----------------------------------

        # ------ reaction data ------------

        # rigid base
        node = inputs["kidx"] + np.ones(len(inputs["kidx"]))  # add one because 0-based index but 1-based node numbering
        rigid = RIGID

        reactions = pyframe3dd.ReactionData(
            node, inputs["kx"], inputs["ky"], inputs["kz"], inputs["ktx"], inputs["kty"], inputs["ktz"], rigid
        )
        # -----------------------------------

        # ------ frame element data ------------
        element = np.arange(1, n)
        N1 = np.arange(1, n)
        N2 = np.arange(2, n + 1)

        roll = np.zeros(n - 1)

        # average across element b.c. frame3dd uses constant section elements
        # TODO: Use nodal2sectional
        Az = inputs["Az"]
        Asx = inputs["Asx"]
        Asy = inputs["Asy"]
        Jz = inputs["Jz"]
        Ixx = inputs["Ixx"]
        Iyy = inputs["Iyy"]
        E = inputs["E"]
        G = inputs["G"]
        rho = inputs["rho"]

        elements = pyframe3dd.ElementData(element, N1, N2, Az, Asx, Asy, Jz, Ixx, Iyy, E, G, roll, rho)
        # -----------------------------------

        # ------ options ------------
        dx = -1.0
        options = pyframe3dd.Options(frame3dd_opt["shear"], frame3dd_opt["geom"], dx)
        # -----------------------------------

        # initialize frame3dd object
        cylinder = pyframe3dd.Frame(nodes, reactions, elements, options)

        # ------ add extra mass ------------

        # extra node inertia data
        N = inputs["midx"] + np.ones(len(inputs["midx"]))

        add_gravity = True
        cylinder.changeExtraNodeMass(
            N,
            inputs["m"],
            inputs["mIxx"],
            inputs["mIyy"],
            inputs["mIzz"],
            inputs["mIxy"],
            inputs["mIxz"],
            inputs["mIyz"],
            inputs["mrhox"],
            inputs["mrhoy"],
            inputs["mrhoz"],
            add_gravity,
        )

        # ------------------------------------

        # ------- enable dynamic analysis ----------
        Mmethod = 1
        lump = 0
        shift = 0.0
        # Run twice the number of modes to ensure that we can ignore the torsional modes and still get the desired number of fore-aft, side-side modes
        cylinder.enableDynamics(2 * NFREQ, Mmethod, lump, frame3dd_opt["tol"], shift)
        # ----------------------------

        # ------ static load case 1 ------------

        # gravity in the X, Y, Z, directions (global)
        gx = 0.0
        gy = 0.0
        gz = -gravity

        load = pyframe3dd.StaticLoadCase(gx, gy, gz)

        # point loads
        nF = inputs["plidx"] + np.ones(len(inputs["plidx"]))
        load.changePointLoads(nF, inputs["Fx"], inputs["Fy"], inputs["Fz"], inputs["Mxx"], inputs["Myy"], inputs["Mzz"])

        # distributed loads
        Px, Py, Pz = inputs["Pz"], inputs["Py"], -inputs["Px"]  # switch to local c.s.
        z = inputs["z"]

        # trapezoidally distributed loads
        EL = np.arange(1, n)
        xx1 = xy1 = xz1 = np.zeros(n - 1)
        xx2 = xy2 = xz2 = np.diff(z) - 1e-6  # subtract small number b.c. of precision
        wx1 = Px[:-1]
        wx2 = Px[1:]
        wy1 = Py[:-1]
        wy2 = Py[1:]
        wz1 = Pz[:-1]
        wz2 = Pz[1:]

        load.changeTrapezoidalLoads(EL, xx1, xx2, wx1, wx2, xy1, xy2, wy1, wy2, xz1, xz2, wz1, wz2)

        cylinder.addLoadCase(load)
        # Debugging
        # cylinder.write('temp.3dd')
        # -----------------------------------
        # run the analysis
        displacements, forces, reactions, internalForces, mass, modal = cylinder.run()
        iCase = 0

        # mass
        outputs["mass"] = mass.struct_mass

        # natural frequncies
        outputs["f1"] = modal.freq[0]
        outputs["f2"] = modal.freq[1]
        outputs["freqs"] = modal.freq[:NFREQ]

        # Get all mode shapes in batch
        NFREQ2 = int(NFREQ / 2)
        freq_x, freq_y, mshapes_x, mshapes_y = util.get_xy_mode_shapes(
            z, modal.freq, modal.xdsp, modal.ydsp, modal.zdsp, modal.xmpf, modal.ympf, modal.zmpf
        )
        outputs["x_mode_freqs"] = freq_x[:NFREQ2]
        outputs["y_mode_freqs"] = freq_y[:NFREQ2]
        outputs["x_mode_shapes"] = mshapes_x[:NFREQ2, :]
        outputs["y_mode_shapes"] = mshapes_y[:NFREQ2, :]

        # deflections due to loading (from cylinder top and wind/wave loads)
        outputs["top_deflection"] = displacements.dx[iCase, n - 1]  # in yaw-aligned direction

        # shear and bending, one per element (convert from local to global c.s.)
        Fz = forces.Nx[iCase, 1::2]
        Vy = forces.Vy[iCase, 1::2]
        Vx = -forces.Vz[iCase, 1::2]

        Mzz = forces.Txx[iCase, 1::2]
        Myy = forces.Myy[iCase, 1::2]
        Mxx = -forces.Mzz[iCase, 1::2]

        # Record total forces and moments
        outputs["base_F"] = -1.0 * np.array([reactions.Fx.sum(), reactions.Fy.sum(), reactions.Fz.sum()])
        outputs["base_M"] = -1.0 * np.array([reactions.Mxx.sum(), reactions.Myy.sum(), reactions.Mzz.sum()])

        outputs["Fz_out"] = Fz
        outputs["Vx_out"] = Vx
        outputs["Vy_out"] = Vy
        outputs["Mxx_out"] = Mxx
        outputs["Myy_out"] = Myy
        outputs["Mzz_out"] = Mzz

        # axial and shear stress
        d, _ = util.nodal2sectional(inputs["d"])
        qdyn, _ = util.nodal2sectional(inputs["qdyn"])

        ##R = self.d/2.0
        ##x_stress = R*np.cos(self.theta_stress)
        ##y_stress = R*np.sin(self.theta_stress)
        ##axial_stress = Fz/self.Az + Mxx/self.Ixx*y_stress - Myy/self.Iyy*x_stress
        #        V = Vy*x_stress/R - Vx*y_stress/R  # shear stress orthogonal to direction x,y
        #        shear_stress = 2. * V / self.Az  # coefficient of 2 for a hollow circular section, but should be conservative for other shapes
        outputs["axial_stress"] = (
            Fz / inputs["Az"] - np.sqrt(Mxx ** 2 + Myy ** 2) / inputs["Iyy"] * d / 2.0
        )  # More conservative, just use the tilted bending and add total max shear as well at the same point, if you do not like it go back to the previous lines

        outputs["shear_stress"] = (
            2.0 * np.sqrt(Vx ** 2 + Vy ** 2) / inputs["Az"]
        )  # coefficient of 2 for a hollow circular section, but should be conservative for other shapes

        # hoop_stress (Eurocode method)
        L_reinforced = self.options["buckling_length"] * np.ones(Fz.shape)
        outputs["hoop_stress_euro"] = hoopStressEurocode(inputs["z"], d, inputs["t"], L_reinforced, qdyn)

        # Simpler hoop stress used in API calculations
        outputs["hoop_stress"] = hoopStress(d, inputs["t"], qdyn)



import numpy as np
import numpy.testing as npt
import unittest
import wisdem.floatingse.map_mooring as mapMooring
from wisdem.pymap import pyMAP
from wisdem.commonse import gravity as g


def myisnumber(instr):
    try:
        float(instr)
    except:
        return False
    return True


myones = np.ones((100,))
truth = [
    "---------------------- LINE DICTIONARY ---------------------------------------",
    "LineType  Diam      MassDenInAir   EA            CB   CIntDamp  Ca   Cdn    Cdt",
    "(-)       (m)       (kg/m)        (N)           (-)   (Pa-s)    (-)  (-)    (-)",
    "chain   0.05    49.75   213500000.0   0.65   1.0E8   0.6   -1.0   0.05",
    "---------------------- NODE PROPERTIES ---------------------------------------",
    "Node Type X     Y    Z   M     V FX FY FZ",
    "(-)  (-) (m)   (m)  (m) (kg) (m^3) (kN) (kN) (kN)",
    "1   VESSEL   11.0   0.0   -10.0   0.0   0.0   #   #   #",
    "2   FIX   175.0   0.0   depth   0.0   0.0   #   #   #",
    "---------------------- LINE PROPERTIES ---------------------------------------",
    "Line    LineType  UnstrLen  NodeAnch  NodeFair  Flags",
    "(-)      (-)       (m)       (-)       (-)       (-)",
    "1   chain   235.8   2   1",
    "---------------------- SOLVER OPTIONS-----------------------------------------",
    "Option",
    "(-)",
    "help",
    " integration_dt 0",
    " kb_default 3.0e6",
    " cb_default 3.0e5",
    " wave_kinematics",
    "inner_ftol 1e-5",
    "inner_gtol 1e-5",
    "inner_xtol 1e-5",
    "outer_tol 1e-3",
    " pg_cooked 10000 1",
    " outer_fd",
    " outer_bd",
    " outer_cd",
    " inner_max_its 200",
    " outer_max_its 600",
    "repeat 120 240",
    " krylov_accelerator 3",
    " ref_position 0.0 0.0 0.0",
]


class TestMapMooring(unittest.TestCase):
    def setUp(self):
        self.inputs = {}
        self.outputs = {}
        self.discrete_inputs = {}

        self.inputs["fairlead"] = 10.0
        self.inputs["fairlead_radius"] = 11.0
        self.inputs["anchor_radius"] = 175.0

        self.inputs["rho_water"] = 1025.0  # 1e3
        self.inputs["water_depth"] = 218.0  # 100.0

        self.inputs["number_of_mooring_connections"] = 3
        self.inputs["mooring_lines_per_connection"] = 1
        self.inputs["mooring_line_length"] = 0.6 * (self.inputs["water_depth"] + self.inputs["anchor_radius"])
        self.inputs["mooring_diameter"] = 0.05
        self.discrete_inputs["mooring_type"] = "chain"
        self.discrete_inputs["anchor_type"] = "suctionpile"
        self.inputs["drag_embedment_extra_length"] = 300.0
        self.inputs["max_offset"] = 10.0
        self.inputs["max_survival_heel"] = 10.0
        self.inputs["operational_heel"] = 10.0
        self.inputs["mooring_cost_factor"] = 1.1

        opt = {}
        opt["gamma_f"] = 1.35

        self.mymap = mapMooring.MapMooring(modeling_options=opt)
        self.mymap.set_properties(self.inputs, self.discrete_inputs)
        self.mymap.set_geometry(self.inputs, self.outputs)

    def testWriteInputAll(self):
        self.mymap.write_input_file(self.inputs, self.discrete_inputs)
        actual = self.mymap.finput[:]
        expect = truth[:]
        self.assertEqual(len(expect), len(actual))

        for n in range(len(actual)):
            actualTok = actual[n].split()
            expectTok = expect[n].split()
            self.assertEqual(len(expectTok), len(actualTok))

            for k in range(len(actualTok)):
                if myisnumber(actualTok[k]):
                    self.assertEqual(float(actualTok[k]), float(expectTok[k]))
                else:
                    self.assertEqual(actualTok[k], expectTok[k])

    def testRunMap(self):
        self.mymap.runMAP(self.inputs, self.discrete_inputs, self.outputs)

        self.assertEqual(np.count_nonzero(self.outputs["mooring_neutral_load"]), 9)
        self.assertEqual(np.count_nonzero(self.outputs["mooring_stiffness"]), 36)
        self.assertEqual(np.count_nonzero(self.outputs["operational_heel_restoring_force"]), 9)
        self.assertGreater(np.count_nonzero(self.outputs["mooring_plot_matrix"]), 9 * 20 - 3)

    def testCost(self):
        self.mymap.compute_cost(self.inputs, self.discrete_inputs, self.outputs)

    def testListEntry(self):
        # Initiate MAP++ for this design
        mymap = pyMAP()
        # mymap.ierr = 0
        mymap.map_set_sea_depth(self.inputs["water_depth"])
        mymap.map_set_gravity(g)
        mymap.map_set_sea_density(self.inputs["rho_water"])
        mymap.read_list_input(truth)
        mymap.init()
        mymap.displace_vessel(0, 0, 0, 0, 10, 0)
        mymap.update_states(0.0, 0)
        mymap.end()


def suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(TestMapMooring))
    return suite


if __name__ == "__main__":
    unittest.TextTestRunner().run(suite())



import numpy as np
import numpy.testing as npt
import unittest
import wisdem.commonse.vertical_cylinder as vc
from wisdem.commonse.utilities import nodal2sectional

npts = 100
myones = np.ones((npts,))


class TestDiscretization(unittest.TestCase):
    def setUp(self):
        self.params = {}
        self.unknowns = {}
        self.resid = None

        self.params["section_height"] = np.arange(1, 5)
        self.params["diameter"] = 5.0 * np.ones(5)
        self.params["wall_thickness"] = 0.05 * np.ones(4)
        self.params["foundation_height"] = 0.0

    def testRefine2(self):
        mydis = vc.CylinderDiscretization(nPoints=5, nRefine=2)
        mydis.compute(self.params, self.unknowns)
        npt.assert_array_equal(self.unknowns["z_param"], np.array([0.0, 1.0, 3.0, 6.0, 10.0]))
        npt.assert_array_equal(self.unknowns["z_full"], np.array([0.0, 0.5, 1.0, 2.0, 3.0, 4.5, 6.0, 8.0, 10.0]))
        npt.assert_array_equal(self.unknowns["d_full"], 5.0)
        npt.assert_array_equal(self.unknowns["t_full"], 0.05)

    def testFoundation(self):
        self.params["foundation_height"] = -30.0
        mydis = vc.CylinderDiscretization(nPoints=5, nRefine=2)
        mydis.compute(self.params, self.unknowns)
        npt.assert_array_equal(self.unknowns["z_param"], np.array([0.0, 1.0, 3.0, 6.0, 10.0]) - 30.0)
        npt.assert_array_equal(self.unknowns["z_full"], np.array([0.0, 0.5, 1.0, 2.0, 3.0, 4.5, 6.0, 8.0, 10.0]) - 30.0)
        npt.assert_array_equal(self.unknowns["d_full"], 5.0)
        npt.assert_array_equal(self.unknowns["t_full"], 0.05)

    def testRefine3(self):
        mydis = vc.CylinderDiscretization(nPoints=5, nRefine=2)
        mydis.compute(self.params, self.unknowns)
        for k in self.unknowns["z_param"]:
            self.assertIn(k, self.unknowns["z_full"])

    def testRefineInterp(self):
        self.params["diameter"] = np.array([5.0, 5.0, 6.0, 7.0, 7.0])
        self.params["wall_thickness"] = 1e-2 * np.array([5.0, 5.0, 6.0, 7.0])
        mydis = vc.CylinderDiscretization(nPoints=5, nRefine=2)
        mydis.compute(self.params, self.unknowns)
        npt.assert_array_equal(self.unknowns["z_param"], np.array([0.0, 1.0, 3.0, 6.0, 10.0]))
        npt.assert_array_equal(self.unknowns["z_full"], np.array([0.0, 0.5, 1.0, 2.0, 3.0, 4.5, 6.0, 8.0, 10.0]))
        npt.assert_array_equal(self.unknowns["d_full"], np.array([5.0, 5.0, 5.0, 5.5, 6.0, 6.5, 7.0, 7.0, 7.0]))
        npt.assert_array_equal(self.unknowns["t_full"], 1e-2 * np.array([5.0, 5.0, 5.0, 5.0, 6.0, 6.0, 7.0, 7.0]))


class TestMass(unittest.TestCase):
    def setUp(self):
        self.params = {}
        self.unknowns = {}
        self.resid = None

        self.params["d_full"] = 2.0 * 10.0 * myones
        self.params["t_full"] = 0.5 * np.ones((npts - 1,))
        self.params["z_full"] = np.linspace(0, 50.0, npts)
        self.params["rho"] = 5.0
        self.params["outfitting_factor"] = 1.5
        self.params["material_cost_rate"] = 1.5
        self.params["labor_cost_rate"] = 1.0
        self.params["painting_cost_rate"] = 10.0

        self.cm = vc.CylinderMass(nPoints=npts)

    def testRegular(self):
        # Straight column
        self.cm.compute(self.params, self.unknowns)

        expect = np.pi * (10.0 ** 2 - 9.5 ** 2) * 5.0 * 1.5 * (50.0 / (npts - 1))
        m = expect * (npts - 1)
        Iax = 0.5 * m * (10.0 ** 2 + 9.5 ** 2)
        Ix = (1 / 12.0) * m * (3 * (10.0 ** 2 + 9.5 ** 2) + 50 * 50) + m * 25 * 25  # parallel axis on last term
        z_avg, _ = nodal2sectional(self.params["z_full"])
        self.assertAlmostEqual(self.unknowns["mass"].sum(), m)
        npt.assert_almost_equal(self.unknowns["mass"], expect)
        npt.assert_almost_equal(self.unknowns["section_center_of_mass"], z_avg)
        self.assertAlmostEqual(self.unknowns["center_of_mass"], 25.0)
        npt.assert_almost_equal(self.unknowns["I_base"], [Ix, Ix, Iax, 0.0, 0.0, 0.0], decimal=5)

        """
    def testFrustum(self):
        # Frustum shell
        self.params['t_full'] = np.array([0.5, 0.4, 0.3])
        self.params['d_full'] = 2*np.array([10.0, 8.0, 6.0])
        self.wave.compute(self.params, self.unknowns)

        expect = np.pi/3.0*5.0*1.5*np.array([20.0, 30.0])*np.array([9.75*1.4+7.8*1.3, 7.8*1.1+5.85*1.0])
        m = expect*(npts-1)
        self.assertAlmostEqual(self.unknowns['mass'].sum(), m)
        npt.assert_almost_equal(self.unknowns['mass'].sum(), expect)
        """


def suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(TestDiscretization))
    suite.addTest(unittest.makeSuite(TestMass))
    return suite


if __name__ == "__main__":
    unittest.TextTestRunner().run(suite())



import openmdao.api as om
from wisdem.floatingse.column import Column
from wisdem.floatingse.substructure import Substructure, SubstructureGeometry
from wisdem.floatingse.loading import Loading
from wisdem.floatingse.map_mooring import MapMooring
from wisdem.towerse.tower import TowerLeanSE
import numpy as np


class FloatingSE(om.Group):
    def initialize(self):
        self.options.declare("modeling_options")

    def setup(self):
        opt = self.options["modeling_options"]["platform"]
        n_mat = self.options["modeling_options"]["materials"]["n_mat"]
        n_height_main = opt["columns"]["main"]["n_height"]
        n_height_off = opt["columns"]["offset"]["n_height"]
        n_height_tow = self.options["modeling_options"]["TowerSE"]["n_height_tower"]

        self.set_input_defaults("mooring_type", "chain")
        self.set_input_defaults("anchor_type", "SUCTIONPILE")
        self.set_input_defaults("loading", "hydrostatic")
        self.set_input_defaults("wave_period_range_low", 2.0, units="s")
        self.set_input_defaults("wave_period_range_high", 20.0, units="s")
        self.set_input_defaults("cd_usr", -1.0)
        self.set_input_defaults("zref", 100.0)
        self.set_input_defaults("number_of_offset_columns", 0)
        self.set_input_defaults("material_names", ["steel"])

        self.add_subsystem(
            "tow",
            TowerLeanSE(modeling_options=self.options["modeling_options"]),
            promotes=[
                "tower_s",
                "tower_height",
                "tower_outer_diameter_in",
                "tower_layer_thickness",
                "tower_outfitting_factor",
                "rna_mass",
                "rna_cg",
                "rna_I",
                "tower_mass",
                "tower_I_base",
                "hub_height",
                "material_names",
                "labor_cost_rate",
                "painting_cost_rate",
                "unit_cost_mat",
                "rho_mat",
                "E_mat",
                "G_mat",
                "sigma_y_mat",
                ("transition_piece_height", "main_freeboard"),
                ("foundation_height", "main_freeboard"),
            ],
        )

        # Next do main and ballast columns
        # Ballast columns are replicated from same design in the components
        column_promotes = [
            "E_mat",
            "G_mat",
            "sigma_y_mat",
            "rho_air",
            "mu_air",
            "rho_water",
            "mu_water",
            "rho_mat",
            "shearExp",
            "yaw",
            "Uc",
            "water_depth",
            "hsig_wave",
            "Tsig_wave",
            "cd_usr",
            "cm",
            "loading",
            "beta_wind",
            "beta_wave",
            "max_draft",
            "material_names",
            "permanent_ballast_density",
            "outfitting_factor",
            "ballast_cost_rate",
            "unit_cost_mat",
            "labor_cost_rate",
            "painting_cost_rate",
            "outfitting_cost_rate",
            "Uref",
            "zref",
            "wind_z0",
        ]
        main_column_promotes = column_promotes.copy()
        main_column_promotes.append(("freeboard", "main_freeboard"))

        self.add_subsystem(
            "main",
            Column(modeling_options=opt, column_options=opt["columns"]["main"], n_mat=n_mat),
            promotes=main_column_promotes,
        )

        off_column_promotes = column_promotes.copy()
        off_column_promotes.append(("freeboard", "off_freeboard"))

        self.add_subsystem(
            "off",
            Column(modeling_options=opt, column_options=opt["columns"]["offset"], n_mat=n_mat),
            promotes=off_column_promotes,
        )

        # Run Semi Geometry for interfaces
        self.add_subsystem(
            "sg", SubstructureGeometry(n_height_main=n_height_main, n_height_off=n_height_off), promotes=["*"]
        )

        # Next run MapMooring
        self.add_subsystem("mm", MapMooring(modeling_options=opt), promotes=["*"])

        # Add in the connecting truss
        self.add_subsystem(
            "load",
            Loading(
                n_height_main=n_height_main, n_height_off=n_height_off, n_height_tow=n_height_tow, modeling_options=opt
            ),
            promotes=["*"],
        )

        # Run main Semi analysis
        self.add_subsystem(
            "subs",
            Substructure(n_height_main=n_height_main, n_height_off=n_height_off, n_height_tow=n_height_tow),
            promotes=["*"],
        )

        # Connect all input variables from all models
        self.connect("tow.d_full", ["windLoads.d", "tower_d_full"])
        self.connect("tow.d_full", "tower_d_base", src_indices=[0])
        self.connect("tow.t_full", "tower_t_full")
        self.connect("tow.z_full", ["loadingWind.z", "windLoads.z", "tower_z_full"])  # includes tower_z_full
        self.connect("tow.E_full", "tower_E_full")
        self.connect("tow.G_full", "tower_G_full")
        self.connect("tow.rho_full", "tower_rho_full")
        self.connect("tow.sigma_y_full", "tower_sigma_y_full")
        self.connect("tow.cm.mass", "tower_mass_section")
        self.connect("tow.turbine_mass", "main.stack_mass_in")
        self.connect("tow.tower_center_of_mass", "tower_center_of_mass")
        self.connect("tow.tower_cost", "tower_shell_cost")

        self.connect("main.z_full", ["main_z_nodes", "main_z_full"])
        self.connect("main.d_full", "main_d_full")
        self.connect("main.t_full", "main_t_full")
        self.connect("main.E_full", "main_E_full")
        self.connect("main.G_full", "main_G_full")
        self.connect("main.rho_full", "main_rho_full")
        self.connect("main.sigma_y_full", "main_sigma_y_full")

        self.connect("off.z_full", ["offset_z_nodes", "offset_z_full"])
        self.connect("off.d_full", "offset_d_full")
        self.connect("off.t_full", "offset_t_full")
        self.connect("off.E_full", "offset_E_full")
        self.connect("off.G_full", "offset_G_full")
        self.connect("off.rho_full", "offset_rho_full")
        self.connect("off.sigma_y_full", "offset_sigma_y_full")

        self.connect("max_offset_restoring_force", "mooring_surge_restoring_force")
        self.connect("operational_heel_restoring_force", "mooring_pitch_restoring_force")

        self.connect("main.z_center_of_mass", "main_center_of_mass")
        self.connect("main.z_center_of_buoyancy", "main_center_of_buoyancy")
        self.connect("main.I_column", "main_moments_of_inertia")
        self.connect("main.Iwater", "main_Iwaterplane")
        self.connect("main.Awater", "main_Awaterplane")
        self.connect("main.displaced_volume", "main_displaced_volume")
        self.connect("main.hydrostatic_force", "main_hydrostatic_force")
        self.connect("main.column_added_mass", "main_added_mass")
        self.connect("main.column_total_mass", "main_mass")
        self.connect("main.column_total_cost", "main_cost")
        self.connect("main.variable_ballast_interp_zpts", "water_ballast_zpts_vector")
        self.connect("main.variable_ballast_interp_radius", "water_ballast_radius_vector")
        self.connect("main.Px", "main_Px")
        self.connect("main.Py", "main_Py")
        self.connect("main.Pz", "main_Pz")
        self.connect("main.qdyn", "main_qdyn")

        self.connect("off.z_center_of_mass", "offset_center_of_mass")
        self.connect("off.z_center_of_buoyancy", "offset_center_of_buoyancy")
        self.connect("off.I_column", "offset_moments_of_inertia")
        self.connect("off.Iwater", "offset_Iwaterplane")
        self.connect("off.Awater", "offset_Awaterplane")
        self.connect("off.displaced_volume", "offset_displaced_volume")
        self.connect("off.hydrostatic_force", "offset_hydrostatic_force")
        self.connect("off.column_added_mass", "offset_added_mass")
        self.connect("off.column_total_mass", "offset_mass")
        self.connect("off.column_total_cost", "offset_cost")
        self.connect("off.Px", "offset_Px")
        self.connect("off.Py", "offset_Py")
        self.connect("off.Pz", "offset_Pz")
        self.connect("off.qdyn", "offset_qdyn")
        self.connect("off.draft", "offset_draft")



import numpy as np
import numpy.testing as npt
import unittest
import wisdem.floatingse.substructure as subs
from wisdem.commonse.vertical_cylinder import get_nfull
from wisdem.commonse import gravity as g

NSECTIONS = 5
NHEIGHT = NSECTIONS + 1
NPTS = get_nfull(NHEIGHT)


class TestSubs(unittest.TestCase):
    def setUp(self):
        self.inputs = {}
        self.outputs = {}
        self.resids = {}

        self.inputs["structural_mass"] = 1e4
        self.inputs["structure_center_of_mass"] = 40.0 * np.ones(3)
        self.inputs["structural_frequencies"] = 100.0 * np.ones(6)
        self.inputs["total_force"] = 26.0 * np.ones(3)
        self.inputs["total_moment"] = 5e4 * np.ones(3)
        self.inputs["total_displacement"] = 1e4

        self.inputs["mooring_mass"] = 20.0
        self.inputs["mooring_neutral_load"] = np.zeros((15, 3))
        self.inputs["mooring_neutral_load"][:3, :] = 5.0 * g
        self.inputs["mooring_surge_restoring_force"] = 1e2
        self.inputs["mooring_pitch_restoring_force"] = 1e5 * np.ones((10, 3))
        self.inputs["mooring_pitch_restoring_force"][3:, :] = 0.0
        self.inputs["mooring_cost"] = 256.0
        self.inputs["mooring_stiffness"] = np.ones((6, 6))
        self.inputs["mooring_moments_of_inertia"] = np.array([10.0, 10.0, 2.0, 0.0, 0.0, 0.0])
        self.inputs["fairlead"] = 0.5
        self.inputs["fairlead_location"] = 0.1
        self.inputs["fairlead_radius"] = 5.0
        self.inputs["max_survival_heel"] = 10.0
        self.inputs["operational_heel"] = 10.0

        self.inputs["pontoon_cost"] = 512.0

        self.inputs["hsig_wave"] = 10.0

        self.inputs["main_Iwaterplane"] = 150.0
        self.inputs["main_Awaterplane"] = 20.0
        self.inputs["main_mass"] = 2.0 * np.ones(NPTS - 1)
        self.inputs["main_cost"] = 32.0
        self.inputs["main_freeboard"] = 10.0
        self.inputs["main_center_of_mass"] = -10.0
        self.inputs["main_center_of_buoyancy"] = -8.0
        self.inputs["main_added_mass"] = 2 * np.array([10.0, 10.0, 2.0, 30.0, 30.0, 0.0])
        self.inputs["main_moments_of_inertia"] = 1e2 * np.array([10.0, 10.0, 2.0, 0.0, 0.0, 0.0])

        self.inputs["offset_Iwaterplane"] = 50.0
        self.inputs["offset_Awaterplane"] = 9.0
        self.inputs["offset_cost"] = 64.0
        self.inputs["offset_mass"] = np.ones(NPTS - 1)
        self.inputs["offset_center_of_mass"] = -5.0
        self.inputs["offset_center_of_buoyancy"] = -4.0
        self.inputs["offset_added_mass"] = np.array([10.0, 10.0, 2.0, 30.0, 30.0, 0.0])
        self.inputs["offset_moments_of_inertia"] = 1e1 * np.array([10.0, 10.0, 2.0, 0.0, 0.0, 0.0])
        self.inputs["offset_freeboard"] = 10.0
        self.inputs["offset_draft"] = 15.0

        self.inputs["tower_z_full"] = np.linspace(0, 90, 3 * NSECTIONS + 1)
        self.inputs["tower_mass"] = 2e2
        self.inputs["tower_shell_cost"] = 2e5
        self.inputs["tower_d_full"] = 5.0 * np.ones(NPTS)
        self.inputs["tower_d_base"] = self.inputs["tower_d_full"][0]
        self.inputs["tower_I_base"] = 1e5 * np.array([1.0, 1.0, 1.0, 0.0, 0.0, 0.0])
        self.inputs["rna_mass"] = 6e1
        self.inputs["rna_cg"] = np.array([0.0, 0.0, 5.0])
        self.inputs["rna_I"] = 1e5 * np.array([1.0, 1.0, 1.0, 0.0, 0.0, 0.0])
        self.inputs["Rhub"] = 3.0

        self.inputs["number_of_offset_columns"] = 3
        self.inputs["water_ballast_radius_vector"] = 40.0 * np.ones(5)
        self.inputs["water_ballast_zpts_vector"] = np.array([-10, -9, -8, -7, -6])
        self.inputs["radius_to_offset_column"] = 20.0
        self.inputs["z_center_of_buoyancy"] = -2.0

        self.inputs["rho_water"] = 1e3
        self.inputs["wave_period_range_low"] = 2.0
        self.inputs["wave_period_range_high"] = 20.0

        self.mysemi = subs.Substructure(n_height_main=NHEIGHT, n_height_off=NHEIGHT, n_height_tow=NHEIGHT)
        self.mysemiG = subs.SubstructureGeometry(n_height_main=NHEIGHT, n_height_off=NHEIGHT)

    def testSetGeometry(self):
        self.inputs["number_of_offset_columns"] = 3
        self.inputs["Rhub"] = 3.0
        self.inputs["tower_d_base"] = 10.0
        self.inputs["main_d_full"] = 2 * np.array([10.0, 10.0, 10.0])
        self.inputs["offset_d_full"] = 2 * np.array([10.0, 10.0, 10.0])
        self.inputs["offset_z_nodes"] = np.array([-35.0, -15.0, 15.0])
        self.inputs["main_z_nodes"] = np.array([-35.0, -15.0, 15.0])
        self.inputs["radius_to_offset_column"] = 25.0
        self.inputs["fairlead_location"] = 0.1
        self.inputs["fairlead_offset_from_shell"] = 1.0
        self.inputs["offset_freeboard"] = 10.0
        self.inputs["offset_draft"] = 15.0
        self.mysemiG.compute(self.inputs, self.outputs)

        # Semi
        self.assertEqual(self.outputs["fairlead"], 30.0)
        self.assertEqual(self.outputs["fairlead_radius"], 11.0 + 25.0)
        self.assertEqual(self.outputs["main_offset_spacing"], 25.0 - 10.0 - 10.0)
        self.assertEqual(self.outputs["tower_transition_buffer"], 10 - 5.0)
        self.assertEqual(self.outputs["offset_freeboard_heel_margin"], 10.0 - 25.0 * np.sin(np.deg2rad(10.0)))
        self.assertEqual(self.outputs["offset_draft_heel_margin"], 15.0 - 25.0 * np.sin(np.deg2rad(10.0)))

        # Spar
        self.inputs["number_of_offset_columns"] = 0
        self.mysemiG.compute(self.inputs, self.outputs)
        self.assertEqual(self.outputs["fairlead"], 30.0)
        self.assertEqual(self.outputs["fairlead_radius"], 11.0)
        self.assertEqual(self.outputs["main_offset_spacing"], 25.0 - 10.0 - 10.0)
        self.assertEqual(self.outputs["tower_transition_buffer"], 10 - 5.0)

    def testBalance(self):
        self.mysemi.balance(self.inputs, self.outputs)
        m_water = 1e3 * 1e4 - 1e4 - 15
        z_data = self.inputs["water_ballast_zpts_vector"]
        h_data = z_data - z_data[0]
        h_expect = np.interp(m_water, 1e3 * h_data * np.pi * self.inputs["water_ballast_radius_vector"] ** 2, h_data)
        cg_expect_z = (1e4 * 40.0 + m_water * (-10 + 0.5 * h_expect)) / (1e4 + m_water)
        cg_expect_xy = 1e4 * 40.0 / (1e4 + m_water)

        self.assertEqual(self.outputs["variable_ballast_mass"], m_water)
        self.assertEqual(self.outputs["variable_ballast_height_ratio"], h_expect / 4.0)
        npt.assert_almost_equal(self.outputs["center_of_mass"], np.array([cg_expect_xy, cg_expect_xy, cg_expect_z]))

        self.inputs["number_of_offset_columns"] = 0
        self.mysemi.balance(self.inputs, self.outputs)

        self.assertEqual(self.outputs["variable_ballast_mass"], m_water)
        self.assertEqual(self.outputs["variable_ballast_height_ratio"], h_expect / 4.0)
        npt.assert_almost_equal(self.outputs["center_of_mass"], np.array([cg_expect_xy, cg_expect_xy, cg_expect_z]))

    def testStability(self):
        self.inputs["mooring_pitch_restoring_force"] = 0.0 * np.ones((10, 3))
        self.outputs["center_of_mass"] = np.array([0.0, 0.0, -1.0])
        self.mysemi.compute_stability(self.inputs, self.outputs)

        I_expect = 150.0 + (50.0 + 9.0 * (20.0 * np.cos(np.deg2rad(np.array([0.0, 120.0, 240.0])))) ** 2).sum()
        static_expect = -1.0 + 2.0
        meta_expect = I_expect / 1e4 - static_expect
        wind_fact = np.cos(np.deg2rad(10.0)) ** 2.0
        self.assertEqual(self.outputs["buoyancy_to_gravity"], static_expect)
        self.assertEqual(self.outputs["metacentric_height"], meta_expect)
        self.assertEqual(self.outputs["offset_force_ratio"], 26.0 / 1e2)
        self.assertAlmostEqual(
            self.outputs["heel_moment_ratio"],
            (wind_fact * 5e4) / (1e4 * g * 1e3 * np.sin(np.deg2rad(10)) * np.abs(meta_expect)),
        )

        self.inputs["number_of_offset_columns"] = 0
        self.mysemi.compute_stability(self.inputs, self.outputs)

        I_expect = 150.0
        meta_expect = I_expect / 1e4 - static_expect
        self.assertEqual(self.outputs["buoyancy_to_gravity"], static_expect)
        self.assertEqual(self.outputs["metacentric_height"], meta_expect)
        self.assertEqual(self.outputs["offset_force_ratio"], 26.0 / 1e2)
        self.assertAlmostEqual(
            self.outputs["heel_moment_ratio"],
            (wind_fact * 5e4) / (1e4 * g * 1e3 * np.sin(np.deg2rad(10)) * np.abs(meta_expect)),
        )

        self.inputs["fairlead"] = 1.0
        self.inputs["mooring_pitch_restoring_force"][:3, -1] = 1.0
        self.assertAlmostEqual(
            self.outputs["heel_moment_ratio"],
            (wind_fact * 5e4) / (1 * 5 + 1e4 * g * 1e3 * np.sin(np.deg2rad(10)) * np.abs(meta_expect)),
        )

    def testPeriods(self):
        # Spar first
        self.inputs["structure_center_of_mass"] = np.array([0.0, 0.0, -40.0])
        self.inputs["number_of_offset_columns"] = 0
        self.mysemi.balance(self.inputs, self.outputs)
        self.inputs["main_center_of_mass"] = self.outputs["center_of_mass"][-1]
        self.inputs["main_center_of_buoyancy"] = self.outputs["center_of_mass"][-1] + 2.0
        self.inputs["tower_mass"] = 0.0
        self.inputs["rna_mass"] = 0.0
        self.inputs["rna_I"] = 1e2 * np.array([1.0, 1.0, 1.0, 0.0, 0.0, 0.0])
        self.inputs["tower_I_base"] = 1e2 * np.array([1.0, 1.0, 1.0, 0.0, 0.0, 0.0])
        self.mysemi.compute_stability(self.inputs, self.outputs)
        self.mysemi.compute_rigid_body_periods(self.inputs, self.outputs)

        m_struct = self.inputs["structural_mass"]
        m_water = self.outputs["variable_ballast_mass"]
        z_cg = self.inputs["main_center_of_mass"]
        z_water = self.outputs["variable_ballast_center_of_mass"]
        I_water = self.outputs["variable_ballast_moments_of_inertia"]
        M_expect = np.zeros(6)
        M_expect[:3] = m_struct + m_water
        M_expect[3:] = self.inputs["main_moments_of_inertia"][:3]
        M_expect[3:5] += I_water[:2] + m_water * (z_water - z_cg) ** 2
        M_expect[-1] += I_water[2]
        M_expect[3:] += 2e2
        npt.assert_equal(self.outputs["mass_matrix"], M_expect)

        A_expect = np.zeros(6)
        A_expect[:3] = self.inputs["main_added_mass"][:3]
        A_expect[3:5] = self.inputs["main_added_mass"][3:5] + A_expect[0] * (2.0) ** 2
        npt.assert_equal(self.outputs["added_mass_matrix"], A_expect)

        rho_w = self.inputs["rho_water"]
        K_expect = np.zeros(6)
        K_expect[2] = rho_w * g * 20.0  # waterplane area for main
        K_expect[3:5] = rho_w * g * self.outputs["metacentric_height"] * 1e4  # Total displacement
        npt.assert_almost_equal(self.outputs["hydrostatic_stiffness"], K_expect)

        T_expect = (
            2 * np.pi * np.sqrt((M_expect + A_expect) / (1e-6 + K_expect + np.diag(self.inputs["mooring_stiffness"])))
        )
        npt.assert_almost_equal(self.outputs["rigid_body_periods"], T_expect)

    def testMargins(self):

        self.mysemi.balance(self.inputs, self.outputs)
        self.mysemi.compute_stability(self.inputs, self.outputs)
        self.mysemi.compute_rigid_body_periods(self.inputs, self.outputs)
        self.mysemi.check_frequency_margins(self.inputs, self.outputs)

        myones = np.ones(6)
        T_sys = self.outputs["rigid_body_periods"]
        T_wave_low = self.inputs["wave_period_range_low"] * myones
        T_wave_high = self.inputs["wave_period_range_high"] * myones
        f_struct = self.inputs["structural_frequencies"]
        T_struct = 1.0 / f_struct

        T_wave_high[-1] = 1e-16
        T_wave_low[-1] = 1e30

        ind = T_sys > T_wave_low
        npt.assert_equal(self.outputs["period_margin_high"][ind], T_sys[ind] / T_wave_high[ind])

        ind = T_sys < T_wave_high
        npt.assert_equal(self.outputs["period_margin_low"][ind], T_sys[ind] / T_wave_low[ind])

        ind = T_struct > T_wave_low
        npt.assert_equal(self.outputs["modal_margin_high"][ind], T_struct[ind] / T_wave_high[ind])

        ind = T_struct < T_wave_high
        npt.assert_equal(self.outputs["modal_margin_low"][ind], T_struct[ind] / T_wave_low[ind])

    def testCost(self):
        self.mysemi.compute_costs(self.inputs, self.outputs)
        c_expect = 256.0 + 512.0 + 32.0 + 3 * 64.0 + 2e5
        self.assertEqual(self.outputs["total_cost"], c_expect)


def suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(TestSubs))
    return suite


if __name__ == "__main__":
    unittest.TextTestRunner().run(suite())



#   Copyright (C) 2014 mdm
#   map[dot]plus[dot]plus[dot]help[at]gmail
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import print_function

from ctypes import *
import os
import six

from distutils.sysconfig import get_config_var

from sys import platform

libext = get_config_var("EXT_SUFFIX")
if libext is None or libext == "":
    if platform == "linux" or platform == "linux2":
        libext = ".so"
    elif platform == "darwin":
        # libext = '.dyld'
        libext = ".so"
    elif platform == "win32":
        # libext = '.dll'
        libext = ".pyd"
    elif platform == "cygwin":
        libext = ".dll"

maplib = "_libmap" + libext

libpath = os.path.dirname(os.path.abspath(__file__)) + os.path.sep + maplib

"""A collection of methods for the MAP++ program"""
libexec = cdll.LoadLibrary(libpath)

"""
# these are the fortran derived types created by the FAST registry.
f_type_init = None
f_type_initout = None
f_type_d = None
f_type_u = None
f_type_x = None
f_type_y = None
f_type_z = None
f_type_p = None

ierr = c_int(0)
status = create_string_buffer(1024)
summary_file = c_char_p
val = c_double
"""


class ModelData_Type(Structure):
    _fields_ = []


# void * object ;
# double gravity ;
# double seaDensity ;
# double depth ;
# char fileName[255] ;
# char summaryFileName[255] ;
# char libraryInputLine[255] ;
# char nodeInputLine[255] ;
# char elementInputLine[255] ;
# char optionInputLine[255] ;
class InitializationData_Type(Structure):
    _fields_ = [
        ("object", c_void_p),
        ("gravity", c_double),
        ("seaDensity", c_double),
        ("depth", c_double),
        ("fileName", c_char * 255),
        ("summaryFileName", c_char * 255),
        ("libraryInputLine", c_char * 255),
        ("nodeInputLine", c_char * 255),
        ("elementInputLine", c_char * 255),
        ("optionInputLine", c_char * 255),
    ]


# void * object ;
# char progName[99] ;
# char version[99] ;
# char compilingData[24] ;
# char * writeOutputHdr ;     int writeOutputHdr_Len ;
# char * writeOutputUnt ;     int writeOutputUnt_Len ;


class InitializationOutputData_Type(Structure):
    _fields_ = [
        ("object", c_void_p),
        ("progName", c_char * 99),
        ("version", c_char * 99),
        ("CompilingData", c_char * 99),
        ("writeOutputHdr", c_char_p),
        ("writeOutputHdr_Len", c_int),
        ("writeOutputUnt", c_char_p),
        ("writeOutputUnt_Len", c_int),
    ]


class InputData_Type(Structure):
    _fields_ = []


class OutputData_Type(Structure):
    _fields_ = []


# void * object ;
# double g ;
# double depth ;
# double rhoSea ;


class ParameterData_Type(Structure):
    _fields_ = [("object", c_void_p), ("g", c_double), ("depth", c_double), ("rhoSea", c_double)]


class ConstraintData_Type(Structure):
    _fields_ = []


class ContinuousData_Type(Structure):
    _fields_ = []


# fields for the fortran types
#
# MAP_EXTERNCALL MAP_InitInputType_t* map_create_init_type( char* msg, MAP_ERROR_CODE* status );
# MAP_EXTERNCALL MAP_InitOutputType_t* map_create_initout_type( char* msg, MAP_ERROR_CODE* status );
# MAP_EXTERNCALL MAP_InputType_t* map_create_input_type( char* msg, MAP_ERROR_CODE* status );
# MAP_EXTERNCALL MAP_ParameterType_t* map_create_parameter_type( char* msg, MAP_ERROR_CODE* status );
# MAP_EXTERNCALL MAP_ConstraintStateType_t* map_create_constraint_type( char* msg, MAP_ERROR_CODE* status );
# MAP_EXTERNCALL MAP_OtherStateType_t* map_create_other_type( char* msg, MAP_ERROR_CODE* status );
# MAP_EXTERNCALL MAP_OutputType_t* map_create_output_type( char* msg, MAP_ERROR_CODE* status );
# MAP_EXTERNCALL MAP_ContinuousStateType_t* map_create_continuous_type( char* msg, MAP_ERROR_CODE* status );

MapData_Type = POINTER(ModelData_Type)
MapInit_Type = POINTER(InitializationData_Type)
MapInitOut_Type = POINTER(InitializationOutputData_Type)
MapInput_Type = POINTER(InputData_Type)
MapOutput_Type = POINTER(OutputData_Type)
MapParameter_Type = POINTER(ParameterData_Type)
MapConstraint_Type = POINTER(ConstraintData_Type)
MapContinuous_Type = POINTER(ContinuousData_Type)

# read file stuff
libexec.set_init_to_null.argtype = [MapInit_Type, c_char_p, POINTER(c_int)]
libexec.map_set_summary_file_name.argtype = [MapInit_Type, c_char_p, POINTER(c_int)]
libexec.map_add_cable_library_input_text.argtype = [MapInit_Type]
libexec.map_add_node_input_text.argtype = [MapInit_Type]
libexec.map_add_line_input_text.argtype = [MapInit_Type]
libexec.map_add_options_input_text.argtype = [MapInit_Type]

libexec.map_create_init_type.argtype = [c_char_p, POINTER(c_int)]
libexec.map_create_initout_type.argtype = [c_char_p, POINTER(c_int)]
libexec.map_create_input_type.argtype = [c_char_p, POINTER(c_int)]
libexec.map_create_parameter_type.argtype = [c_char_p, POINTER(c_int)]
libexec.map_create_constraint_type.argtype = [c_char_p, POINTER(c_int)]
libexec.map_create_other_type.argtype = [c_char_p, POINTER(c_int)]
libexec.map_create_output_type.argtype = [c_char_p, POINTER(c_int)]
libexec.map_create_continuous_type.argtype = [c_char_p, POINTER(c_int)]
libexec.map_create_continuous_type.argtype = [MapData_Type]

libexec.map_create_init_type.restype = MapInit_Type
libexec.map_create_initout_type.restype = MapInitOut_Type
libexec.map_create_input_type.restype = MapInput_Type
libexec.map_create_parameter_type.restype = MapParameter_Type
libexec.map_create_constraint_type.restype = MapConstraint_Type
libexec.map_create_other_type.restype = MapData_Type
libexec.map_create_output_type.restype = MapOutput_Type
libexec.map_create_continuous_type.restype = MapContinuous_Type


libexec.map_set_sea_depth.argtypes = [MapParameter_Type, c_double]
libexec.map_set_gravity.argtypes = [MapParameter_Type, c_double]
libexec.map_set_sea_density.argtypes = [MapParameter_Type, c_double]

libexec.map_size_lines.restype = c_int

# numeric routines
libexec.map_residual_function_length.restype = c_double
libexec.map_residual_function_height.restype = c_double
libexec.map_jacobian_dxdh.restype = c_double
libexec.map_jacobian_dxdv.restype = c_double
libexec.map_jacobian_dzdh.restype = c_double
libexec.map_jacobian_dzdv.restype = c_double

libexec.map_residual_function_length.argtypes = [MapData_Type, c_int, c_char_p, POINTER(c_int)]
libexec.map_residual_function_height.argtypes = [MapData_Type, c_int, c_char_p, POINTER(c_int)]
libexec.map_jacobian_dxdh.argtypes = [MapData_Type, c_int, c_char_p, POINTER(c_int)]
libexec.map_jacobian_dxdv.argtypes = [MapData_Type, c_int, c_char_p, POINTER(c_int)]
libexec.map_jacobian_dzdh.argtypes = [MapData_Type, c_int, c_char_p, POINTER(c_int)]
libexec.map_jacobian_dzdv.argtypes = [MapData_Type, c_int, c_char_p, POINTER(c_int)]


libexec.map_get_fairlead_force_2d.argtypes = [
    POINTER(c_double),
    POINTER(c_double),
    MapData_Type,
    c_int,
    c_char_p,
    POINTER(c_int),
]
libexec.map_get_anchor_force_2d.argtypes = [
    POINTER(c_double),
    POINTER(c_double),
    MapData_Type,
    c_int,
    c_char_p,
    POINTER(c_int),
]

libexec.map_get_fairlead_force_3d.argtypes = [
    POINTER(c_double),
    POINTER(c_double),
    POINTER(c_double),
    MapData_Type,
    c_int,
    c_char_p,
    POINTER(c_int),
]
libexec.map_get_anchor_force_3d.argtypes = [
    POINTER(c_double),
    POINTER(c_double),
    POINTER(c_double),
    MapData_Type,
    c_int,
    c_char_p,
    POINTER(c_int),
]


# plot routines
libexec.map_plot_x_array.argtypes = [MapData_Type, c_int, c_int, c_char_p, POINTER(c_int)]
libexec.map_plot_x_array.restype = POINTER(c_double)
libexec.map_plot_y_array.argtypes = [MapData_Type, c_int, c_int, c_char_p, POINTER(c_int)]
libexec.map_plot_y_array.restype = POINTER(c_double)
libexec.map_plot_z_array.argtypes = [MapData_Type, c_int, c_int, c_char_p, POINTER(c_int)]
libexec.map_plot_z_array.restype = POINTER(c_double)
libexec.map_plot_array_free.argtypes = [POINTER(c_double)]


# modifyers
libexec.map_offset_vessel.argtypes = [
    MapData_Type,
    MapInput_Type,
    c_double,
    c_double,
    c_double,
    c_double,
    c_double,
    c_double,
    c_char_p,
    POINTER(c_int),
]
libexec.map_linearize_matrix.argtypes = [
    MapInput_Type,
    MapParameter_Type,
    MapData_Type,
    MapOutput_Type,
    MapConstraint_Type,
    c_double,
    POINTER(c_int),
    c_char_p,
]
libexec.map_linearize_matrix.restype = POINTER(POINTER(c_double))
libexec.map_free_linearize_matrix.argtypes = [POINTER(POINTER(c_double))]

libexec.map_init.argtypes = [
    MapInit_Type,
    MapInput_Type,
    MapParameter_Type,
    MapContinuous_Type,
    MapConstraint_Type,
    MapData_Type,
    MapOutput_Type,
    MapInitOut_Type,
    POINTER(c_int),
    c_char_p,
]


libexec.map_update_states.argtypes = [
    c_float,
    c_int,
    MapInput_Type,
    MapParameter_Type,
    MapContinuous_Type,
    MapConstraint_Type,
    MapData_Type,
    POINTER(c_int),
    c_char_p,
]

libexec.map_end.argtypes = [
    MapInput_Type,
    MapParameter_Type,
    MapContinuous_Type,
    MapConstraint_Type,
    MapData_Type,
    MapOutput_Type,
    POINTER(c_int),
    c_char_p,
]


libexec.map_initialize_msqs_base.argtypes = [
    MapInput_Type,
    MapParameter_Type,
    MapContinuous_Type,
    MapConstraint_Type,
    MapData_Type,
    MapOutput_Type,
    MapInitOut_Type,
]


libexec.map_size_lines.argtypes = [MapData_Type, POINTER(c_int), c_char_p]


class pyMAP(object):
    def __init__(self):
        self.ierr = c_int(0)
        self.status = create_string_buffer(1024)

        self.f_type_d = self.CreateDataState()
        self.f_type_u = self.CreateInputState()
        self.f_type_x = self.CreateContinuousState()
        self.f_type_p = self.CreateParameterState()
        self.f_type_y = self.CreateOutputState()
        self.f_type_z = self.CreateConstraintState()
        self.f_type_init = self.CreateInitState()
        self.f_type_initout = self.CreateInitoutState()
        libexec.set_init_to_null(self.f_type_init, self.status, pointer(self.ierr))
        libexec.map_initialize_msqs_base(
            self.f_type_u,
            self.f_type_p,
            self.f_type_x,
            self.f_type_z,
            self.f_type_d,
            self.f_type_y,
            self.f_type_initout,
        )
        self.summary_file("outlist.map.sum")

    def init(self):
        libexec.map_init(
            self.f_type_init,
            self.f_type_u,
            self.f_type_p,
            self.f_type_x,
            self.f_type_z,
            self.f_type_d,
            self.f_type_y,
            self.f_type_initout,
            pointer(self.ierr),
            self.status,
        )
        if self.ierr.value != 0:
            print(self.status.value)

    def size_lines(self):
        size = libexec.map_size_lines(self.f_type_d, pointer(self.ierr), self.status)
        if self.ierr.value != 0:
            print(self.status.value)
        return size

    def update_states(self, t, interval):
        libexec.map_update_states(
            c_float(t),
            c_int(interval),
            self.f_type_u,
            self.f_type_p,
            self.f_type_x,
            self.f_type_z,
            self.f_type_d,
            pointer(self.ierr),
            self.status,
        )
        if self.ierr.value != 0:
            print(self.status.value)

    # Calls function in main.c and fordatamanager.c to delete insteads of c structs. First, the malloc'ed arrays need to vanish
    # gracefully; we accomplish this by calling MAP_End(...) routine. Then, the structs themself are deleted. Order is important.
    #
    # MAP_EXTERNCALL int MAP_End ( InputData *u, ParameterData *p, ContinuousData *x, ConstraintData *z, ModelData *data, OutputData *y, char *map_msg, MAP_ERROR_CODE *ierr )
    # MAP_EXTERNCALL void MAP_Input_Delete( InputData* u )
    # MAP_EXTERNCALL void MAP_Param_Delete( ParameterData* p )
    # MAP_EXTERNCALL void MAP_ContState_Delete( InputData* x )
    # MAP_EXTERNCALL void MAP_ConstrState_Delete( InputData* z )
    # MAP_EXTERNCALL void MAP_Output_Delete( InputData* y )
    # MAP_EXTERNCALL void MAP_OtherState_Delete( ModelData* data )
    def end(self):
        libexec.map_end(
            self.f_type_u,
            self.f_type_p,
            self.f_type_x,
            self.f_type_z,
            self.f_type_d,
            self.f_type_y,
            pointer(self.ierr),
            self.status,
        )

    # Set a name for the MAP summary file. Does not need to be called. If not called, the default name is 'outlist.sum.map'
    def summary_file(self, echo_file):
        self.f_type_init.contents.summaryFileName = six.b(echo_file)
        libexec.map_set_summary_file_name(self.f_type_init, self.status, pointer(self.ierr))

    # Calls function in fortdatamanager.c to create instance of c structs
    # MAP_EXTERNCALL InitializationData* MAP_InitInput_Create( char* map_msg, MAP_ERROR_CODE* ierr )
    def CreateInitState(self):
        obj = libexec.map_create_init_type(self.status, pointer(self.ierr))
        if self.ierr.value != 0:
            print(self.status.value)
        return obj

    # Calls function in fortdatamanager.c to create instance of c structs
    # MAP_EXTERNCALL void MAP_InitOutput_Delete( InputData* io )
    def CreateInitoutState(self):
        obj = libexec.map_create_initout_type(self.status, pointer(self.ierr))
        if self.ierr.value != 0:
            print(self.status.value)
        return obj

    # Calls function in fortdatamanager.c to create instance of c structs
    # MAP_EXTERNCALL ModelData *MAP_OtherState_Create( char *map_msg, MAP_ERROR_CODE *ierr )
    def CreateDataState(self):
        obj = libexec.map_create_other_type(self.status, pointer(self.ierr))
        if self.ierr.value != 0:
            print(self.status.value)
        return obj

    # Calls function in fortdatamanager.c to create instance of c structs
    # MAP_EXTERNCALL InputData* MAP_Input_Create( char* map_msg, MAP_ERROR_CODE *ierr )
    def CreateInputState(self):
        obj = libexec.map_create_input_type(self.status, pointer(self.ierr))
        if self.ierr.value != 0:
            print(self.status.value)
        return obj

    # Calls function in fortdatamanager.c to create instance of c structs
    # MAP_EXTERNCALL ContinuousData* MAP_ContState_Create( char* map_msg, MAP_ERROR_CODE *ierr )
    def CreateContinuousState(self):
        obj = libexec.map_create_continuous_type(self.status, pointer(self.ierr))
        if self.ierr.value != 0:
            print(self.status.value)
        return obj

    # Calls function in fortdatamanager.c to create instance of c structs
    # MAP_EXTERNCALL OutputData *MAP_Output_Create( char *map_msg, MAP_ERROR_CODE *ierr )
    def CreateOutputState(self):
        obj = libexec.map_create_output_type(self.status, pointer(self.ierr))
        if self.ierr.value != 0:
            print(self.status.value)
        return obj

    # Calls function in fortdatamanager.c to create instance of c structs
    # MAP_EXTERNCALL ConstraintData* MAP_ConstrState_Create( char* map_msg, MAP_ERROR_CODE *ierr )
    def CreateConstraintState(self):
        obj = libexec.map_create_constraint_type(self.status, pointer(self.ierr))
        if self.ierr.value != 0:
            print(self.status.value)
        return obj

    # Calls function in fortdatamanager.c to create instance of c structs
    # MAP_EXTERNCALL ParameterData* MAP_Param_Create( char* map_msg, MAP_ERROR_CODE *ierr )
    def CreateParameterState(self):
        obj = libexec.map_create_parameter_type(self.status, pointer(self.ierr))
        if self.ierr.value != 0:
            print(self.status.value)
        return obj

    def map_set_sea_depth(self, depth):
        libexec.map_set_sea_depth(self.f_type_p, depth)

    def map_set_gravity(self, g):
        libexec.map_set_gravity(self.f_type_p, g)

    def map_set_sea_density(self, rho):
        libexec.map_set_sea_density(self.f_type_p, rho)

    def plot_x(self, lineNum, length):
        arr = [None] * length
        array = POINTER(c_double)
        array = libexec.map_plot_x_array(self.f_type_d, lineNum, length, self.status, pointer(self.ierr))
        if self.ierr.value != 0:
            print(self.status.value)
            self.end()
            libexec.map_plot_array_free(array)
            raise RuntimeError("MAP terminated premature.")
        arr = [array[j] for j in range(length)]
        libexec.map_plot_array_free(array)
        return arr

    def plot_y(self, lineNum, length):
        arr = [None] * length
        array = POINTER(c_double)
        array = libexec.map_plot_y_array(self.f_type_d, lineNum, length, self.status, pointer(self.ierr))
        if self.ierr.value != 0:
            print(self.status.value)
            self.end()
            libexec.map_plot_array_free(array)
            raise RuntimeError("MAP terminated premature.")
        arr = [array[j] for j in range(length)]
        libexec.map_plot_array_free(array)
        return arr

    def plot_z(self, lineNum, length):
        arr = [None] * length
        array = POINTER(c_double)
        array = libexec.map_plot_z_array(self.f_type_d, lineNum, length, self.status, pointer(self.ierr))
        if self.ierr.value != 0:
            print(self.status.value)
            self.end()
            libexec.map_plot_array_free(array)
            raise RuntimeError("MAP terminated premature.")
        arr = [array[j] for j in range(length)]
        libexec.map_plot_array_free(array)
        return arr

    def get_fairlead_force_2d(self, index):
        """Gets the horizontal and vertical fairlead force in a 2D plane along the
        straight-line line. Must ensure update_states() is called before accessing
        this function. The function will not solve the forces for a new vessel position
        if it updated. , otherwise the fairlead forces are not updated with the new
        vessel position. Called C function:

        MAP_EXTERNCALL void map_get_fairlead_force_2d(double* H, double* V, MAP_OtherStateType_t* other_type, int index, char* map_msg, MAP_ERROR_CODE* ierr);

        :param index: The line number the fairlead forces are being requested for. Zero indexed
        :returns: horizontal and vertical fairlead force [N]

        >>> H,V = print get_fairlead_force_2d(1)
        """
        H_ref = c_double(-999.9)
        V_ref = c_double(-999.9)
        libexec.map_get_fairlead_force_2d(
            pointer(H_ref), pointer(V_ref), self.f_type_d, index, self.status, pointer(self.ierr)
        )
        return H_ref.value, V_ref.value

    def get_fairlead_force_3d(self, index):
        """Gets the horizontal and vertical fairlead force in a 3D frame along relative
        referene global axis. Must ensure update_states() is called before accessing
        this function. The function will not solve the forces for a new vessel position
        if it updated. , otherwise the fairlead forces are not updated with the new
        vessel position. Called C function:

        MAP_EXTERNCALL void map_get_fairlead_force_3d(double* fx, double* fy, double* fz, MAP_OtherStateType_t* other_type, int index, char* map_msg, MAP_ERROR_CODE* ierr);

        :param index: The line number the fairlead forces are being requested for. Zero indexed
        :returns: horizontal and vertical fairlead force [N]

        >>> fx,fy,fz = get_fairlead_force_3d(1)
        """
        fx = c_double(-999.9)
        fy = c_double(-999.9)
        fz = c_double(-999.9)
        libexec.map_get_fairlead_force_3d(
            pointer(fx), pointer(fy), pointer(fz), self.f_type_d, index, self.status, pointer(self.ierr)
        )
        return fx.value, fy.value, fz.value

    def get_anchor_force_2d(self, index):
        """Gets the horizontal and vertical anchor force in a 2D plane along the
        straight-line line. Must ensure update_states() is called before accessing
        this function. The function will not solve the forces for a new vessel position
        if it updated. , otherwise the anchor forces are not updated with the new
        vessel position. Called C function:

        MAP_EXTERNCALL void map_get_anchor_force_2d(double* H, double* V, MAP_OtherStateType_t* other_type, int index, char* map_msg, MAP_ERROR_CODE* ierr);

        :param index: The line number the anchor forces are being requested for. Zero indexed
        :returns: horizontal and vertical anchor force [N]

        >>> H,V = print get_anchor_force_2d(1)
        """
        Ha_ref = c_double(-999.9)
        Va_ref = c_double(-999.9)
        libexec.map_get_anchor_force_2d(
            pointer(Ha_ref), pointer(Va_ref), self.f_type_d, index, self.status, pointer(self.ierr)
        )
        return Ha_ref.value, Va_ref.value

    def get_anchor_force_3d(self, index):
        """Gets the horizontal and vertical anchor force in a 3D frame along relative
        referene global axis. Must ensure update_states() is called before accessing
        this function. The function will not solve the forces for a new vessel position
        if it updated. , otherwise the anchor forces are not updated with the new
        vessel position. Called C function:

        MAP_EXTERNCALL void map_get_anchor_force_3d(double* fx, double* fy, double* fz, MAP_OtherStateType_t* other_type, int index, char* map_msg, MAP_ERROR_CODE* ierr);

        :param index: The line number the anchor forces are being requested for. Zero indexed
        :returns: horizontal and vertical anchor force [N]

        >>> fx,fy,fz = get_anchor_force_3d(1)
        """
        fxa = c_double(-999.9)
        fya = c_double(-999.9)
        fza = c_double(-999.9)
        libexec.map_get_anchor_force_3d(
            pointer(fxa), pointer(fya), pointer(fza), self.f_type_d, index, self.status, pointer(self.ierr)
        )
        return fxa.value, fya.value, fza.value

    def funcl(self, i):
        self.val = libexec.map_residual_function_length(self.f_type_d, i, self.status, pointer(self.ierr))
        if self.ierr.value != 0:
            print(self.status.value)
            self.end()
            raise RuntimeError("MAP terminated premature.")
        return self.val

    def funch(self, i):
        self.val = libexec.map_residual_function_height(self.f_type_d, i, self.status, pointer(self.ierr))
        if self.ierr.value != 0:
            print(self.status.value)
            self.end()
            raise RuntimeError("MAP terminated premature.")
        return self.val

    def dxdh(self, i):
        self.val = libexec.map_jacobian_dxdh(self.f_type_d, i, self.status, pointer(self.ierr))
        if self.ierr.value != 0:
            print(self.status.value)
            self.end()
            raise RuntimeError("MAP terminated premature.")
        return self.val

    def dxdv(self, i):
        self.val = libexec.map_jacobian_dxdv(self.f_type_d, i, self.status, pointer(self.ierr))
        if self.ierr.value != 0:
            print(self.status.value)
            self.end()
            raise RuntimeError("MAP terminated premature.")
        return self.val

    def dzdh(self, i):
        self.val = libexec.map_jacobian_dzdh(self.f_type_d, i, self.status, pointer(self.ierr))
        if self.ierr.value != 0:
            print(self.status.value)
            self.end()
            raise RuntimeError("MAP terminated premature.")
        return self.val

    def dzdv(self, i):
        self.val = libexec.map_jacobian_dzdv(self.f_type_d, i, self.status, pointer(self.ierr))
        if self.ierr.value != 0:
            print(self.status.value)
            self.end()
            raise RuntimeError("MAP terminated premature.")
        return self.val

    def linear(self, epsilon):
        """Insert a function and its arguments in process pool.

        Input is inserted in queues using a round-robin fashion. Every job is
        identified by and index that is returned by function. Not all parameters
        of original multiprocessing.Pool.apply_aync are implemented so far.

        :param func: Function to process.
        :type func: Callable.
        :param args: Arguments for the function to process.
        :type args: Tuple.
        :returns: Assigned job id.
        :rtype: Int.
        """
        array = POINTER(POINTER(c_double))
        array = libexec.map_linearize_matrix(
            self.f_type_u,
            self.f_type_p,
            self.f_type_d,
            self.f_type_y,
            self.f_type_z,
            epsilon,
            pointer(self.ierr),
            self.status,
        )
        if self.ierr.value != 0:
            print(self.status.value)
            self.end()
            raise RuntimeError("MAP terminated premature.")
        arr = [[array[j][i] for i in range(6)] for j in range(6)]
        libexec.map_free_linearize_matrix(array)
        return arr

    def displace_vessel(self, x, y, z, phi, the, psi):
        libexec.map_offset_vessel(self.f_type_d, self.f_type_u, x, y, z, phi, the, psi, self.status, pointer(self.ierr))
        if self.ierr.value != 0:
            print(self.status.value)
            self.end()
            raise RuntimeError("MAP terminated premature.")

    def read_file(self, fileName):
        f = open(fileName, "r")
        charptr = POINTER(c_char)
        line_offset = []
        temp_str = []
        offset = 0

        for line in f:
            line_offset.append(offset)
            offset += len(line)
        f.seek(0)

        i = 0
        for line in f:
            words = line.split()
            if words[0] == "LineType":
                next(f)
                LineType_ref = i
            elif words[0] == "Node":
                next(f)
                Node_ref = i
            elif words[0] == "Line":
                next(f)
                Line_ref = i
            elif words[0] == "Option":
                next(f)
                Option_ref = i
            i += 1

        f.seek(line_offset[LineType_ref + 2])
        for line in f:
            if line[0] == "-":
                break
            else:
                # create_string_buffer(line, 255).raw
                self.f_type_init.contents.libraryInputLine = six.b(line + "\0")
                libexec.map_add_cable_library_input_text(self.f_type_init)

        f.seek(line_offset[Node_ref + 3])
        for line in f:
            if line[0] == "-":
                break
            else:
                self.f_type_init.contents.nodeInputLine = six.b(line + "\0")
                libexec.map_add_node_input_text(self.f_type_init)

        f.seek(line_offset[Line_ref + 4])
        for line in f:
            if line[0] == "-":
                break
            else:
                self.f_type_init.contents.elementInputLine = six.b(line + "\0")
                libexec.map_add_line_input_text(self.f_type_init)

        f.seek(line_offset[Option_ref + 5])
        for line in f:
            if line[0] == "-":
                break
            elif line[0] == "!":
                None
            else:
                self.f_type_init.contents.optionInputLine = six.b(line + "\0")
                libexec.map_add_options_input_text(self.f_type_init)

    def read_list_input(self, listIn):
        assert isinstance(listIn, list), "Must input a python list of strings"
        assert len(listIn) >= 8, "Must have at least 4 sections, 3 lines per section"
        assert type(listIn[0]) == type(""), "List elements must be strings"
        charptr = POINTER(c_char)

        dictFlag = nodeFlag = propFlag = solvFlag = False
        listIter = iter(listIn)
        for line in listIter:
            if "LINE DICTIONARY" in line.upper():
                dictFlag = True
                nodeFlag = propFlag = solvFlag = False
                for _ in range(2):
                    next(listIter)  # Process header
                continue
            elif "NODE PROPERTIES" in line.upper():
                nodeFlag = True
                dictFlag = propFlag = solvFlag = False
                for _ in range(2):
                    next(listIter)  # Process header
                continue
            elif "LINE PROPERTIES" in line.upper():
                propFlag = True
                dictFlag = nodeFlag = solvFlag = False
                for _ in range(2):
                    next(listIter)  # Process header
                continue
            elif "SOLVER OPTIONS" in line.upper():
                solvFlag = True
                dictFlag = nodeFlag = propFlag = False
                for _ in range(2):
                    next(listIter)  # Process header
                continue

            if dictFlag:
                self.f_type_init.contents.libraryInputLine = six.b(line + "\n\0")
                libexec.map_add_cable_library_input_text(self.f_type_init)

            if nodeFlag:
                self.f_type_init.contents.nodeInputLine = six.b(line + "\n\0")
                libexec.map_add_node_input_text(self.f_type_init)

            if propFlag:
                self.f_type_init.contents.elementInputLine = six.b(line + "\n\0")
                libexec.map_add_line_input_text(self.f_type_init)

            if solvFlag:
                self.f_type_init.contents.optionInputLine = six.b(line + "\n\0")
                libexec.map_add_options_input_text(self.f_type_init)



import os

import numpy as np
import openmdao.api as om


class PoseOptimization(object):
    def __init__(self, modeling_options, analysis_options):
        self.modeling = modeling_options
        self.opt = analysis_options

    def get_number_design_variables(self):
        # Determine the number of design variables
        n_DV = 0

        rotorD_opt = self.opt["optimization_variables"]["rotor_diameter"]
        blade_opt = self.opt["optimization_variables"]["blade"]
        tower_opt = self.opt["optimization_variables"]["tower"]
        mono_opt = self.opt["optimization_variables"]["monopile"]
        hub_opt = self.opt["optimization_variables"]["hub"]
        drive_opt = self.opt["optimization_variables"]["drivetrain"]

        if rotorD_opt["flag"]:
            n_DV += 1
        if blade_opt["aero_shape"]["twist"]["flag"]:
            n_DV += blade_opt["aero_shape"]["twist"]["n_opt"] - 2
        if blade_opt["aero_shape"]["chord"]["flag"]:
            n_DV += blade_opt["aero_shape"]["chord"]["n_opt"] - 3
        if blade_opt["aero_shape"]["af_positions"]["flag"]:
            n_DV += self.modeling["RotorSE"]["n_af_span"] - blade_opt["aero_shape"]["af_positions"]["af_start"] - 1
        if blade_opt["structure"]["spar_cap_ss"]["flag"]:
            n_DV += blade_opt["structure"]["spar_cap_ss"]["n_opt"] - 2
        if (
            blade_opt["structure"]["spar_cap_ps"]["flag"]
            and not blade_opt["structure"]["spar_cap_ps"]["equal_to_suction"]
        ):
            n_DV += blade_opt["structure"]["spar_cap_ps"]["n_opt"] - 2
        if self.opt["optimization_variables"]["control"]["tsr"]["flag"]:
            n_DV += 1
        # if self.opt["optimization_variables"]["control"]["servo"]["pitch_control"]["flag"]:
        #    n_DV += 2
        # if self.opt["optimization_variables"]["control"]["servo"]["torque_control"]["flag"]:
        #    n_DV += 2
        if tower_opt["outer_diameter"]["flag"]:
            n_DV += self.modeling["TowerSE"]["n_height_tower"]
        if tower_opt["layer_thickness"]["flag"]:
            n_DV += (self.modeling["TowerSE"]["n_height_tower"] - 1) * self.modeling["TowerSE"]["n_layers_tower"]
        if mono_opt["outer_diameter"]["flag"]:
            n_DV += self.modeling["TowerSE"]["n_height_monopile"]
        if mono_opt["layer_thickness"]["flag"]:
            n_DV += (self.modeling["TowerSE"]["n_height_monopile"] - 1) * self.modeling["TowerSE"]["n_layers_monopile"]
        if hub_opt["cone"]["flag"]:
            n_DV += 1
        if hub_opt["hub_diameter"]["flag"]:
            n_DV += 1
        for k in [
            "uptilt",
            "overhang",
            "distance_tt_hub",
            "distance_hub_mb",
            "distance_mb_mb",
            "generator_length",
            "gear_ratio",
            "generator_length",
            "bedplate_web_thickness",
            "bedplate_flange_thickness",
            "bedplate_flange_width",
        ]:
            if drive_opt[k]["flag"]:
                n_DV += 1
        for k in [
            "lss_diameter",
            "lss_wall_thickness",
            "hss_diameter",
            "hss_wall_thickness",
            "nose_diameter",
            "nose_wall_thickness",
        ]:
            if drive_opt[k]["flag"]:
                n_DV += 2
        if drive_opt["bedplate_wall_thickness"]["flag"]:
            n_DV += 4
        if self.opt["driver"]["form"] == "central":
            n_DV *= 2

        return n_DV

    def _get_step_size(self):
        # If a step size for the driver-level finite differencing is provided, use that step size. Otherwise use a default value.
        return 1.0e-6 if not "step_size" in self.opt["driver"] else self.opt["driver"]["step_size"]

    def set_driver(self, wt_opt):
        folder_output = self.opt["general"]["folder_output"]

        step_size = self._get_step_size()

        # Solver has specific meaning in OpenMDAO
        wt_opt.model.approx_totals(method="fd", step=step_size, form=self.opt["driver"]["form"])

        # Set optimization solver and options. First, Scipy's SLSQP
        if self.opt["driver"]["solver"] == "SLSQP":
            wt_opt.driver = om.ScipyOptimizeDriver()
            wt_opt.driver.options["optimizer"] = self.opt["driver"]["solver"]
            wt_opt.driver.options["tol"] = self.opt["driver"]["tol"]
            wt_opt.driver.options["maxiter"] = self.opt["driver"]["max_iter"]

        # The next two optimization methods require pyOptSparse.
        elif self.opt["driver"]["solver"] == "CONMIN":
            try:
                from openmdao.api import pyOptSparseDriver
            except:
                raise ImportError(
                    "You requested the optimization solver CONMIN, but you have not installed the pyOptSparseDriver. Please do so and rerun."
                )
            wt_opt.driver = pyOptSparseDriver()
            wt_opt.driver.options["optimizer"] = self.opt["driver"]["solver"]
            wt_opt.driver.opt_settings["ITMAX"] = self.opt["driver"]["max_iter"]

        elif self.opt["driver"]["solver"] == "SNOPT":
            try:
                from openmdao.api import pyOptSparseDriver
            except:
                raise ImportError(
                    "You requested the optimization solver SNOPT, but you have not installed the pyOptSparseDriver. Please do so and rerun."
                )
            wt_opt.driver = pyOptSparseDriver()
            try:
                wt_opt.driver.options["optimizer"] = self.opt["driver"]["solver"]
            except:
                raise ImportError(
                    "You requested the optimization solver SNOPT, but you have not installed it within the pyOptSparseDriver. Please do so and rerun."
                )
            wt_opt.driver.opt_settings["Major optimality tolerance"] = float(self.opt["driver"]["tol"])
            wt_opt.driver.opt_settings["Major iterations limit"] = int(self.opt["driver"]["max_major_iter"])
            wt_opt.driver.opt_settings["Iterations limit"] = int(self.opt["driver"]["max_minor_iter"])
            wt_opt.driver.opt_settings["Major feasibility tolerance"] = float(self.opt["driver"]["tol"])
            wt_opt.driver.opt_settings["Summary file"] = os.path.join(folder_output, "SNOPT_Summary_file.txt")
            wt_opt.driver.opt_settings["Print file"] = os.path.join(folder_output, "SNOPT_Print_file.txt")
            if "hist_file_name" in self.opt["driver"]:
                wt_opt.driver.hist_file = self.opt["driver"]["hist_file_name"]
            if "verify_level" in self.opt["driver"]:
                wt_opt.driver.opt_settings["Verify level"] = self.opt["driver"]["verify_level"]
            # wt_opt.driver.declare_coloring()
            if "hotstart_file" in self.opt["driver"]:
                wt_opt.driver.hotstart_file = self.opt["driver"]["hotstart_file"]

        else:
            raise ValueError("The optimizer " + self.opt["driver"]["solver"] + "is not yet supported!")

        return wt_opt

    def set_objective(self, wt_opt):

        # Set merit figure. Each objective has its own scaling.
        if self.opt["merit_figure"] == "AEP":
            wt_opt.model.add_objective("rp.AEP", ref=-1.0e6)

        elif self.opt["merit_figure"] == "blade_mass":
            wt_opt.model.add_objective("re.precomp.blade_mass", ref=1.0e4)

        elif self.opt["merit_figure"] == "LCOE":
            wt_opt.model.add_objective("financese.lcoe", ref=0.1)

        elif self.opt["merit_figure"] == "blade_tip_deflection":
            wt_opt.model.add_objective("tcons.tip_deflection_ratio")

        elif self.opt["merit_figure"] == "tower_mass":
            wt_opt.model.add_objective("towerse.tower_mass", ref=1e6)

        elif self.opt["merit_figure"] == "mononpile_mass":
            wt_opt.model.add_objective("towerse.mononpile_mass", ref=1e6)

        elif self.opt["merit_figure"] == "structural_mass":
            wt_opt.model.add_objective("towerse.structural_mass", ref=1e6)

        elif self.opt["merit_figure"] == "tower_cost":
            wt_opt.model.add_objective("tcc.tower_cost", ref=1e6)

        elif self.opt["merit_figure"] == "hub_mass":
            wt_opt.model.add_objective("drivese.hub_system_mass", ref=1e5)

        elif self.opt["merit_figure"] == "nacelle_mass":
            wt_opt.model.add_objective("drivese.nacelle_mass", ref=1e6)

        elif self.opt["merit_figure"] == "Cp":
            if self.modeling["flags"]["blade"]:
                wt_opt.model.add_objective("rp.powercurve.Cp_regII", ref=-1.0)
            else:
                wt_opt.model.add_objective("ccblade.CP", ref=-1.0)
        else:
            raise ValueError("The merit figure " + self.opt["merit_figure"] + " is not supported.")

        return wt_opt

    def set_design_variables(self, wt_opt, wt_init):

        # Set optimization design variables.
        rotorD_opt = self.opt["optimization_variables"]["rotor_diameter"]
        blade_opt = self.opt["optimization_variables"]["blade"]
        tower_opt = self.opt["optimization_variables"]["tower"]
        monopile_opt = self.opt["optimization_variables"]["monopile"]
        control_opt = self.opt["optimization_variables"]["control"]
        hub_opt = self.opt["optimization_variables"]["hub"]
        drive_opt = self.opt["optimization_variables"]["drivetrain"]

        if rotorD_opt["flag"]:
            wt_opt.model.add_design_var(
                "configuration.rotor_diameter_user", lower=rotorD_opt["minimum"], upper=rotorD_opt["minimum"], ref=1.0e2
            )

        if blade_opt["aero_shape"]["twist"]["flag"]:
            indices = range(2, blade_opt["aero_shape"]["twist"]["n_opt"])
            wt_opt.model.add_design_var("blade.opt_var.twist_opt_gain", indices=indices, lower=0.0, upper=1.0)

        chord_options = blade_opt["aero_shape"]["chord"]
        if chord_options["flag"]:
            indices = range(3, chord_options["n_opt"] - 1)
            wt_opt.model.add_design_var(
                "blade.opt_var.chord_opt_gain",
                indices=indices,
                lower=chord_options["min_gain"],
                upper=chord_options["max_gain"],
            )

        if blade_opt["aero_shape"]["af_positions"]["flag"]:
            n_af = self.modeling["RotorSE"]["n_af_span"]
            indices = range(blade_opt["aero_shape"]["af_positions"]["af_start"], n_af - 1)
            af_pos_init = wt_init["components"]["blade"]["outer_shape_bem"]["airfoil_position"]["grid"]
            step_size = self._get_step_size()
            lb_af = np.zeros(n_af)
            ub_af = np.zeros(n_af)
            for i in range(1, indices[0]):
                lb_af[i] = ub_af[i] = af_pos_init[i]
            for i in indices:
                lb_af[i] = 0.5 * (af_pos_init[i - 1] + af_pos_init[i]) + step_size
                ub_af[i] = 0.5 * (af_pos_init[i + 1] + af_pos_init[i]) - step_size
            lb_af[-1] = ub_af[-1] = 1.0
            wt_opt.model.add_design_var(
                "blade.opt_var.af_position", indices=indices, lower=lb_af[indices], upper=ub_af[indices]
            )

        spar_cap_ss_options = blade_opt["structure"]["spar_cap_ss"]
        if spar_cap_ss_options["flag"]:
            indices = range(1, spar_cap_ss_options["n_opt"] - 1)
            wt_opt.model.add_design_var(
                "blade.opt_var.spar_cap_ss_opt_gain",
                indices=indices,
                lower=spar_cap_ss_options["min_gain"],
                upper=spar_cap_ss_options["max_gain"],
            )

        # Only add the pressure side design variables if we do set
        # `equal_to_suction` as False in the optimization yaml.
        spar_cap_ps_options = blade_opt["structure"]["spar_cap_ps"]
        if spar_cap_ps_options["flag"] and not spar_cap_ps_options["equal_to_suction"]:
            indices = range(1, spar_cap_ps_options["n_opt"] - 1)
            wt_opt.model.add_design_var(
                "blade.opt_var.spar_cap_ps_opt_gain",
                indices=indices,
                lower=spar_cap_ps_options["min_gain"],
                upper=spar_cap_ps_options["max_gain"],
            )

        # -- Tower & Monopile --
        if tower_opt["outer_diameter"]["flag"]:
            wt_opt.model.add_design_var(
                "tower.diameter",
                lower=tower_opt["outer_diameter"]["lower_bound"],
                upper=tower_opt["outer_diameter"]["upper_bound"],
                ref=5.0,
            )

        if tower_opt["layer_thickness"]["flag"]:
            wt_opt.model.add_design_var(
                "tower.layer_thickness",
                lower=tower_opt["layer_thickness"]["lower_bound"],
                upper=tower_opt["layer_thickness"]["upper_bound"],
                ref=1e-2,
            )

        if monopile_opt["outer_diameter"]["flag"]:
            wt_opt.model.add_design_var(
                "monopile.diameter",
                lower=monopile_opt["outer_diameter"]["lower_bound"],
                upper=monopile_opt["outer_diameter"]["upper_bound"],
                ref=5.0,
            )

        if monopile_opt["layer_thickness"]["flag"]:
            wt_opt.model.add_design_var(
                "monopile.layer_thickness",
                lower=monopile_opt["layer_thickness"]["lower_bound"],
                upper=monopile_opt["layer_thickness"]["upper_bound"],
                ref=1e-2,
            )

        # -- Control --
        if control_opt["tsr"]["flag"]:
            wt_opt.model.add_design_var(
                "control.rated_TSR", lower=control_opt["tsr"]["minimum"], upper=control_opt["tsr"]["maximum"], ref=1e1
            )
        # -- Hub & Drivetrain --
        if hub_opt["cone"]["flag"]:
            wt_opt.model.add_design_var(
                "hub.cone", lower=hub_opt["cone"]["lower_bound"], upper=hub_opt["cone"]["upper_bound"], ref=1e-2
            )
        if hub_opt["hub_diameter"]["flag"]:
            wt_opt.model.add_design_var(
                "hub.diameter",
                lower=hub_opt["hub_diameter"]["lower_bound"],
                upper=hub_opt["hub_diameter"]["upper_bound"],
            )
        if drive_opt["uptilt"]["flag"]:
            wt_opt.model.add_design_var(
                "nacelle.uptilt",
                lower=drive_opt["uptilt"]["lower_bound"],
                upper=drive_opt["uptilt"]["upper_bound"],
                ref=1e-2,
            )

        if drive_opt["generator_length"]["flag"]:
            wt_opt.model.add_design_var(
                "nacelle.L_generator",
                lower=drive_opt["generator_length"]["lower_bound"],
                upper=drive_opt["generator_length"]["upper_bound"],
            )

        for k in [
            "overhang",
            "distance_tt_hub",
            "distance_hub_mb",
            "distance_mb_mb",
            "gear_ratio",
            "bedplate_flange_width",
            "lss_diameter",
            "hss_diameter",
            "nose_diameter",
        ]:
            if drive_opt[k]["flag"]:
                wt_opt.model.add_design_var(
                    "nacelle." + k, lower=drive_opt[k]["lower_bound"], upper=drive_opt[k]["upper_bound"]
                )

        for k in [
            "bedplate_wall_thickness",
            "bedplate_web_thickness",
            "bedplate_flange_thickness",
            "lss_wall_thickness",
            "hss_wall_thickness",
            "nose_wall_thickness",
        ]:
            if drive_opt[k]["flag"]:
                wt_opt.model.add_design_var(
                    "nacelle." + k, lower=drive_opt[k]["lower_bound"], upper=drive_opt[k]["upper_bound"], ref=1e-2
                )

        return wt_opt

    def set_constraints(self, wt_opt):
        blade_opt = self.opt["optimization_variables"]["blade"]

        # Set non-linear blade constraints
        blade_constr = self.opt["constraints"]["blade"]
        if blade_constr["strains_spar_cap_ss"]["flag"]:
            if blade_opt["structure"]["spar_cap_ss"]["flag"]:
                wt_opt.model.add_constraint("rs.constr.constr_max_strainU_spar", upper=1.0)
            else:
                print(
                    "WARNING: the strains of the suction-side spar cap are set to be constrained, but spar cap thickness is not an active design variable. The constraint is not enforced."
                )

        if blade_constr["strains_spar_cap_ps"]["flag"]:
            if (
                blade_opt["structure"]["spar_cap_ps"]["flag"]
                or blade_opt["structure"]["spar_cap_ps"]["equal_to_suction"]
            ):
                wt_opt.model.add_constraint("rs.constr.constr_max_strainL_spar", upper=1.0)
            else:
                print(
                    "WARNING: the strains of the pressure-side spar cap are set to be constrained, but spar cap thickness is not an active design variable. The constraint is not enforced."
                )

        if blade_constr["stall"]["flag"]:
            if blade_opt["aero_shape"]["twist"]["flag"]:
                wt_opt.model.add_constraint("stall_check.no_stall_constraint", upper=1.0)
            else:
                print(
                    "WARNING: the margin to stall is set to be constrained, but twist is not an active design variable. The constraint is not enforced."
                )

        if blade_constr["tip_deflection"]["flag"]:
            if blade_opt["structure"]["spar_cap_ss"]["flag"] or blade_opt["structure"]["spar_cap_ps"]["flag"]:
                wt_opt.model.add_constraint("tcons.tip_deflection_ratio", upper=1.0)
            else:
                print(
                    "WARNING: the tip deflection is set to be constrained, but spar caps thickness is not an active design variable. The constraint is not enforced."
                )

        if blade_constr["chord"]["flag"]:
            if blade_opt["aero_shape"]["chord"]["flag"]:
                wt_opt.model.add_constraint("blade.pa.max_chord_constr", upper=1.0)
            else:
                print(
                    "WARNING: the max chord is set to be constrained, but chord is not an active design variable. The constraint is not enforced."
                )

        if blade_constr["frequency"]["flap_3P"]:
            if blade_opt["structure"]["spar_cap_ss"]["flag"] or blade_opt["structure"]["spar_cap_ps"]["flag"]:
                wt_opt.model.add_constraint("rs.constr.constr_flap_f_margin", upper=0.0)
            else:
                print(
                    "WARNING: the blade flap frequencies are set to be constrained, but spar caps thickness is not an active design variable. The constraint is not enforced."
                )

        if blade_constr["frequency"]["edge_3P"]:
            wt_opt.model.add_constraint("rs.constr.constr_edge_f_margin", upper=0.0)

        if blade_constr["rail_transport"]["8_axle"]:
            wt_opt.model.add_constraint("re.rail.constr_LV_8axle_horiz", lower=0.8, upper=1.0)
            wt_opt.model.add_constraint("re.rail.constr_strainPS", upper=1.0)
            wt_opt.model.add_constraint("re.rail.constr_strainSS", upper=1.0)
        elif blade_constr["rail_transport"]["4_axle"]:
            wt_opt.model.add_constraint("re.rail.constr_LV_4axle_horiz", upper=1.0)
            wt_opt.model.add_constraint("re.rail.constr_strainPS", upper=1.0)
            wt_opt.model.add_constraint("re.rail.constr_strainSS", upper=1.0)

        if self.opt["constraints"]["blade"]["moment_coefficient"]["flag"]:
            wt_opt.model.add_constraint(
                "ccblade.CM",
                lower=self.opt["constraints"]["blade"]["moment_coefficient"]["min"],
                upper=self.opt["constraints"]["blade"]["moment_coefficient"]["max"],
            )
        if (
            self.opt["constraints"]["blade"]["match_cl_cd"]["flag_cl"]
            or self.opt["constraints"]["blade"]["match_cl_cd"]["flag_cd"]
        ):
            data_target = np.loadtxt(self.opt["constraints"]["blade"]["match_cl_cd"]["filename"])
            eta_opt = np.linspace(0.0, 1.0, self.opt["optimization_variables"]["blade"]["aero_shape"]["twist"]["n_opt"])
            target_cl = np.interp(eta_opt, data_target[:, 0], data_target[:, 3])
            target_cd = np.interp(eta_opt, data_target[:, 0], data_target[:, 4])
            eps_cl = 1.0e-2
            if self.opt["constraints"]["blade"]["match_cl_cd"]["flag_cl"]:
                wt_opt.model.add_constraint("ccblade.cl_n_opt", lower=target_cl - eps_cl, upper=target_cl + eps_cl)
            if self.opt["constraints"]["blade"]["match_cl_cd"]["flag_cd"]:
                wt_opt.model.add_constraint("ccblade.cd_n_opt", lower=target_cd - eps_cl, upper=target_cd + eps_cl)
        if (
            self.opt["constraints"]["blade"]["match_L_D"]["flag_L"]
            or self.opt["constraints"]["blade"]["match_L_D"]["flag_D"]
        ):
            data_target = np.loadtxt(self.opt["constraints"]["blade"]["match_L_D"]["filename"])
            eta_opt = np.linspace(0.0, 1.0, self.opt["optimization_variables"]["blade"]["aero_shape"]["twist"]["n_opt"])
            target_L = np.interp(eta_opt, data_target[:, 0], data_target[:, 7])
            target_D = np.interp(eta_opt, data_target[:, 0], data_target[:, 8])
        eps_L = 1.0e2
        if self.opt["constraints"]["blade"]["match_L_D"]["flag_L"]:
            wt_opt.model.add_constraint("ccblade.L_n_opt", lower=target_L - eps_L, upper=target_L + eps_L)
        if self.opt["constraints"]["blade"]["match_L_D"]["flag_D"]:
            wt_opt.model.add_constraint("ccblade.D_n_opt", lower=target_D - eps_L, upper=target_D + eps_L)

        # Tower and monopile contraints
        tower_constr = self.opt["constraints"]["tower"]
        monopile_constr = self.opt["constraints"]["monopile"]
        if tower_constr["height_constraint"]["flag"] or monopile_constr["height_constraint"]["flag"]:
            wt_opt.model.add_constraint(
                "towerse.height_constraint",
                lower=tower_constr["height_constraint"]["lower_bound"],
                upper=tower_constr["height_constraint"]["upper_bound"],
            )

        if tower_constr["stress"]["flag"] or monopile_constr["stress"]["flag"]:
            for k in range(self.modeling["TowerSE"]["nLC"]):
                kstr = "" if self.modeling["TowerSE"]["nLC"] == 0 else str(k + 1)
                wt_opt.model.add_constraint("towerse.post" + kstr + ".stress", upper=1.0)

        if tower_constr["global_buckling"]["flag"] or monopile_constr["global_buckling"]["flag"]:
            for k in range(self.modeling["TowerSE"]["nLC"]):
                kstr = "" if self.modeling["TowerSE"]["nLC"] == 0 else str(k + 1)
                wt_opt.model.add_constraint("towerse.post" + kstr + ".global_buckling", upper=1.0)

        if tower_constr["shell_buckling"]["flag"] or monopile_constr["shell_buckling"]["flag"]:
            for k in range(self.modeling["TowerSE"]["nLC"]):
                kstr = "" if self.modeling["TowerSE"]["nLC"] == 0 else str(k + 1)
                wt_opt.model.add_constraint("towerse.post" + kstr + ".shell_buckling", upper=1.0)

        if tower_constr["d_to_t"]["flag"] or monopile_constr["d_to_t"]["flag"]:
            wt_opt.model.add_constraint(
                "towerse.constr_d_to_t",
                lower=tower_constr["d_to_t"]["lower_bound"],
                upper=tower_constr["d_to_t"]["upper_bound"],
            )

        if tower_constr["taper"]["flag"] or monopile_constr["taper"]["flag"]:
            wt_opt.model.add_constraint("towerse.constr_taper", lower=tower_constr["taper"]["lower_bound"])

        if tower_constr["slope"]["flag"] or monopile_constr["slope"]["flag"]:
            wt_opt.model.add_constraint("towerse.slope", upper=1.0)

        if tower_constr["frequency"]["flag"]:
            wt_opt.model.add_constraint("tcons.constr_tower_f_1Pmargin", upper=0.0)
            wt_opt.model.add_constraint("tcons.constr_tower_f_NPmargin", upper=0.0)

        elif tower_constr["frequency_1"]["flag"] or monopile_constr["frequency_1"]["flag"]:
            for k in range(self.modeling["TowerSE"]["nLC"]):
                kstr = "" if self.modeling["TowerSE"]["nLC"] == 0 else str(k + 1)
                wt_opt.model.add_constraint(
                    "towerse.post" + kstr + ".structural_frequencies",
                    indices=[0],
                    lower=tower_constr["frequency_1"]["lower_bound"],
                    upper=tower_constr["frequency_1"]["upper_bound"],
                )

        # Hub and drivetrain constraints
        hub_constr = self.opt["constraints"]["hub"]
        drive_constr = self.opt["constraints"]["drivetrain"]

        if hub_constr["hub_diameter"]["flag"]:
            wt_opt.model.add_constraint("drivese.constr_hub_diameter", lower=0.0)

        for k in ["lss", "hss", "bedplate", "mb1", "mb2"]:
            if drive_constr[k]["flag"]:
                mystr = k + "_vonmises" if k in ["lss", "hss", "bedplate"] else k + "_defl"
                wt_opt.model.add_constraint("drivese.constr_" + mystr, upper=1.0)

        for k in ["length", "height", "access", "ecc"]:
            if drive_constr[k]["flag"]:
                wt_opt.model.add_constraint("drivese.constr_" + k, lower=0.0)

        return wt_opt

    def set_recorders(self, wt_opt):
        folder_output = self.opt["general"]["folder_output"]

        # Set recorder on the OpenMDAO driver level using the `optimization_log`
        # filename supplied in the optimization yaml
        if self.opt["recorder"]["flag"]:
            recorder = om.SqliteRecorder(os.path.join(folder_output, self.opt["recorder"]["file_name"]))
            wt_opt.driver.add_recorder(recorder)
            wt_opt.add_recorder(recorder)

            wt_opt.driver.recording_options["excludes"] = ["*_df"]
            wt_opt.driver.recording_options["record_constraints"] = True
            wt_opt.driver.recording_options["record_desvars"] = True
            wt_opt.driver.recording_options["record_objectives"] = True

        return wt_opt

    def set_initial(self, wt_opt, wt_init):
        blade_opt = self.opt["optimization_variables"]["blade"]

        if self.modeling["flags"]["blade"]:
            wt_opt["blade.opt_var.s_opt_twist"] = np.linspace(0.0, 1.0, blade_opt["aero_shape"]["twist"]["n_opt"])
            if blade_opt["aero_shape"]["twist"]["flag"]:
                init_twist_opt = np.interp(
                    wt_opt["blade.opt_var.s_opt_twist"],
                    wt_init["components"]["blade"]["outer_shape_bem"]["twist"]["grid"],
                    wt_init["components"]["blade"]["outer_shape_bem"]["twist"]["values"],
                )
                lb_twist = np.array(blade_opt["aero_shape"]["twist"]["lower_bound"])
                ub_twist = np.array(blade_opt["aero_shape"]["twist"]["upper_bound"])
                wt_opt["blade.opt_var.twist_opt_gain"] = (init_twist_opt - lb_twist) / (ub_twist - lb_twist)
                if (
                    max(wt_opt["blade.opt_var.twist_opt_gain"]) > 1.0
                    or min(wt_opt["blade.opt_var.twist_opt_gain"]) < 0.0
                ):
                    print(
                        "Warning: the initial twist violates the upper or lower bounds of the twist design variables."
                    )

            blade_constr = self.opt["constraints"]["blade"]
            wt_opt["blade.opt_var.s_opt_chord"] = np.linspace(0.0, 1.0, blade_opt["aero_shape"]["chord"]["n_opt"])
            wt_opt["blade.ps.s_opt_spar_cap_ss"] = np.linspace(0.0, 1.0, blade_opt["structure"]["spar_cap_ss"]["n_opt"])
            wt_opt["blade.ps.s_opt_spar_cap_ps"] = np.linspace(0.0, 1.0, blade_opt["structure"]["spar_cap_ps"]["n_opt"])
            wt_opt["rs.constr.max_strainU_spar"] = blade_constr["strains_spar_cap_ss"]["max"]
            wt_opt["rs.constr.max_strainL_spar"] = blade_constr["strains_spar_cap_ps"]["max"]
            wt_opt["stall_check.stall_margin"] = blade_constr["stall"]["margin"] * 180.0 / np.pi
            wt_opt["tcons.max_allowable_td_ratio"] = blade_constr["tip_deflection"]["margin"]

        if self.modeling["flags"]["nacelle"] and self.modeling["DriveSE"]["direct"]:
            drive_constr = self.opt["constraints"]["drivetrain"]
            wt_opt["drivese.access_diameter"] = drive_constr["access"]["lower_bound"]

        return wt_opt

    def set_restart(self, wt_opt):
        if "warmstart_file" in self.opt["driver"]:

            # Directly read the pyoptsparse sqlite db file
            from pyoptsparse import SqliteDict

            db = SqliteDict(self.opt["driver"]["warmstart_file"])

            # Grab the last iteration's design variables
            last_key = db["last"]
            desvars = db[last_key]["xuser"]

            # Obtain the already-setup OM problem's design variables
            if wt_opt.model._static_mode:
                design_vars = wt_opt.model._static_design_vars
            else:
                design_vars = wt_opt.model._design_vars

            # Get the absolute names from the promoted names within the OM model.
            # We need this because the pyoptsparse db has the absolute names for
            # variables but the OM model uses the promoted names.
            prom2abs = wt_opt.model._var_allprocs_prom2abs_list["output"]
            abs2prom = {}
            for key in design_vars:
                abs2prom[prom2abs[key][0]] = key

            # Loop through each design variable
            for key in desvars:
                prom_key = abs2prom[key]

                # Scale each DV based on the OM scaling from the problem.
                # This assumes we're running the same problem with the same scaling
                scaler = design_vars[prom_key]["scaler"]
                adder = design_vars[prom_key]["adder"]

                if scaler is None:
                    scaler = 1.0
                if adder is None:
                    adder = 0.0

                scaled_dv = desvars[key] / scaler - adder

                # Special handling for blade twist as we only have the
                # last few control points as design variables
                if "twist_opt_gain" in key:
                    wt_opt[key][2:] = scaled_dv
                else:
                    wt_opt[key][:] = scaled_dv

        return wt_opt
